from pathlib import Path
from typing import Any, List, Optional, override

from pydantic import BaseModel, Field, RootModel, model_validator

from poexy_core.packages.files import FilePattern, PackageFiles, ResolvePackageFiles
from poexy_core.packages.format import PackageFormat
from poexy_core.packages.validators import validate_destination
from poexy_core.pyproject.types import GlobPattern

# pylint: disable=no-member


class Exclude(BaseModel, ResolvePackageFiles):
    path: GlobPattern = Field(
        description="Path to exclude from the built archive",
    )
    formats: Optional[List[PackageFormat]] = Field(
        default=None,
        description="List of package formats to exclude. If not specified, all formats "
        "will be excluded.",
    )

    @override
    def resolve(
        self, _format: PackageFormat, base_path: Path
    ) -> Optional[PackageFiles]:
        if self.formats is not None and _format not in self.formats:
            return None
        path, glob_pattern = self.path.split()
        file_pattern = FilePattern(glob_pattern=glob_pattern, path=path)
        if _format == PackageFormat.Source:
            destination_path = path
        else:
            destination_path = base_path
        resolved = file_pattern.resolve(destination_path)
        return resolved


class Include(BaseModel, ResolvePackageFiles):
    path: GlobPattern = Field(description="Path to include in the built archive")
    destination: Optional[Path] = Field(
        description="Relative path to the include path in the built archive",
        default=None,
    )
    formats: Optional[List[PackageFormat]] = Field(
        default=None,
        description="List of package formats to include. If not specified, all formats "
        "will be included.",
    )

    @model_validator(mode="after")
    def validate_model(self) -> "Include":
        if self.destination is not None:
            validate_destination("destination", self.destination)
        return self

    @override
    def resolve(
        self, _format: PackageFormat, base_path: Path
    ) -> Optional[PackageFiles]:
        if self.formats is not None and _format not in self.formats:
            return None
        path, glob_pattern = self.path.split()
        file_pattern = FilePattern(glob_pattern=glob_pattern, path=path)
        if _format == PackageFormat.Source:
            destination_path = path
        elif self.destination is None:
            destination_path = base_path / path
        else:
            destination_path = self.destination / base_path / path
        resolved = file_pattern.resolve(destination_path)
        return resolved


IncludesType = List[Include]
ExcludesType = List[Exclude]


class Includes(RootModel[IncludesType], ResolvePackageFiles):
    @model_validator(mode="before")
    @classmethod
    def validate_model(cls, data: Any) -> Any:
        if not isinstance(data, list):
            return data
        for index, item in enumerate(data):
            if isinstance(item, str):
                data[index] = Include(path=GlobPattern(pattern=Path(item)))
            else:
                data[index] = Include(
                    path=GlobPattern(pattern=Path(item.get("path"))),
                    destination=item.get("destination", None),
                    formats=item.get("formats", None),
                )
        return data

    @override
    def resolve(
        self, _format: PackageFormat, base_path: Path
    ) -> Optional[PackageFiles]:
        resolved = []
        for item in self.root:
            resolved_item = item.resolve(_format, base_path)
            if resolved_item is not None:
                resolved.extend(resolved_item)
        return set(resolved)


class Excludes(RootModel[ExcludesType], ResolvePackageFiles):
    @model_validator(mode="before")
    @classmethod
    def validate_model(cls, data: Any) -> Any:
        if not isinstance(data, list):
            return data
        for index, item in enumerate(data):
            if isinstance(item, str):
                data[index] = Exclude(path=GlobPattern(pattern=Path(item)))
            else:
                data[index] = Exclude(
                    path=GlobPattern(pattern=Path(item.get("path"))),
                    formats=item.get("formats", None),
                )
        return data

    @override
    def resolve(
        self, _format: PackageFormat, base_path: Path
    ) -> Optional[PackageFiles]:
        resolved = []
        for item in self.root:
            resolved_item = item.resolve(_format, base_path)
            if resolved_item is not None:
                resolved.extend(resolved_item)
        return set(resolved)



import logging
from typing import override

from poetry.core.masonry.metadata import Metadata

from poexy_core.builders.hooks.hook import HookBuilder
from poexy_core.builders.types import FilePathCallback, FilePathPredicate
from poexy_core.packages.format import PackageFormat
from poexy_core.pyproject.tables.poexy import Poexy

logger = logging.getLogger(__name__)


class PackageFilesHookBuilder(HookBuilder):

    def __init__(
        self,
        poexy: Poexy,
        _format: PackageFormat,
        destination_predicate: FilePathPredicate,
    ):
        super().__init__("package_files")
        self.__poexy = poexy
        self.__format = _format
        self.__destination_predicate = destination_predicate
        self.__files = []

    @override
    def build(self):
        with self._hook_build():
            logger.info("Resolving package files...")
            resolved = self.__poexy.resolve_package_files(self.__format)
            inclusions = self.__poexy.resolve_inclusions(self.__format)

            logger.info(f"Resolved {len(resolved)} package files.")

            count = 0

            for file in resolved:
                if file.source in inclusions.excludes:
                    logger.info(f"Excluding file: {file.source}")
                    continue
                destination = self.__destination_predicate(file.destination)
                if destination is None:
                    logger.info(f"Excluding file: {file.source}")
                    continue
                logger.info(f"Including file: {file.source}")
                self.__files.append((file.source, destination / file.destination))
                count += 1

            logger.info(f"{count} files ready to be packaged.")

    @override
    def add_files(self, callback: FilePathCallback):
        with self._hook_add_files():
            for source, destination in self.__files:
                callback(source, destination)

    @override
    def add_metadata(self, metadata: Metadata, _format: PackageFormat):
        pass



import os
import sys
import uuid
from pathlib import Path

import pytest

# pylint: disable=redefined-outer-name


@pytest.fixture(scope="session")
def samples_path():
    return Path(__file__).parent.parent / "samples"


@pytest.fixture(scope="session")
def self_project_dist_path(tmpdir_factory) -> Path:
    tmpdir = tmpdir_factory.mktemp("dist")
    return Path(tmpdir)


@pytest.fixture(scope="function")
def tmp_root(tmpdir_factory):
    path = tmpdir_factory.mktemp(f"test_{uuid.uuid4().hex[:8]}")
    return Path(path)


@pytest.fixture(scope="function")
def dist_path(tmp_root):
    return tmp_root / "dist"


@pytest.fixture(scope="function")
def dist_temp_path(dist_path):
    return dist_path / "temp"


@pytest.fixture(scope="function")
def build_path(tmp_root):
    return tmp_root / "build"


@pytest.fixture(scope="function")
def install_path(tmp_root):
    return tmp_root / "install"


@pytest.fixture(scope="session")
def global_virtualenv_path(tmpdir_factory) -> Path:
    tmpdir = tmpdir_factory.mktemp("venv")
    return Path(tmpdir)


@pytest.fixture(scope="session")
def global_virtualenv_archive_path(tmpdir_factory) -> Path:
    tmpdir = tmpdir_factory.mktemp("venv-archive")
    return Path(tmpdir)


@pytest.fixture(scope="function")
def virtualenv_path(tmp_root):
    return tmp_root / "venv"


@pytest.fixture(scope="function")
def pyinstaller_path(tmp_root):
    path = tmp_root / "pyinstaller"

    def _pyinstaller_path():
        os.environ["PYINSTALLER_CONFIG_DIR"] = str(path)

    return _pyinstaller_path


@pytest.fixture(scope="function")
def site_packages_path(install_path) -> Path:
    python_version = f"{sys.version_info.major}.{sys.version_info.minor}"
    path = install_path / "lib" / f"python{python_version}" / "site-packages"
    return path



from typing import Callable

import pytest
from assertpy import assert_that

from poexy_core.builders.builder import PythonTag
from poexy_core.manifest.manifest import (
    MetadataManifest,
    PackageInfoManifest,
    RecordManifest,
    WheelManifest,
)
from poexy_core.packages.files import FORBIDDEN_DIRS


@pytest.fixture()
def assert_metadata_manifest(
    wheel_metadata, package_name, package_version
) -> Callable[[PythonTag], MetadataManifest]:
    def _assert(python_tag: PythonTag):
        metadata = wheel_metadata(python_tag)
        metadata_manifest = MetadataManifest(metadata.dist_info_folder)
        metadata_manifest.read()
        assert_that(metadata_manifest.get("Metadata-Version")).is_equal_to("2.4")
        assert_that(metadata_manifest.get("Name")).is_equal_to(package_name())
        assert_that(metadata_manifest.get("Version")).is_equal_to(package_version())
        return metadata_manifest

    return _assert


@pytest.fixture()
def assert_wheel_manifest(wheel_metadata) -> Callable[[PythonTag], WheelManifest]:
    def _assert(python_tag: PythonTag):
        metadata = wheel_metadata(python_tag)
        wheel_manifest = WheelManifest(metadata.dist_info_folder)
        wheel_manifest.read()
        assert_that(wheel_manifest.get("Wheel-Version")).is_equal_to("1.0")
        assert_that(wheel_manifest.get("Generator")).is_equal_to("Poexy")
        if python_tag.platform is None:
            assert_that(wheel_manifest.get("Root-Is-Purelib")).is_equal_to("true")
        else:
            assert_that(wheel_manifest.get("Root-Is-Purelib")).is_equal_to("false")
        assert_that(wheel_manifest.get("Tag")).is_equal_to(f"{python_tag}")
        return wheel_manifest

    return _assert


@pytest.fixture()
def assert_record_manifest(wheel_metadata) -> Callable[[PythonTag], RecordManifest]:
    def _assert(python_tag: PythonTag):
        metadata = wheel_metadata(python_tag)
        record_manifest = RecordManifest(metadata.dist_info_folder)
        record_manifest.read()
        assert_that(len(record_manifest)).is_greater_than(1)
        for record in record_manifest:
            assert_that(FORBIDDEN_DIRS).does_not_contain(*record.path.parts)
        return record_manifest

    return _assert


@pytest.fixture()
def assert_pkginfo_manifest(
    sdist_metadata, package_name, package_version
) -> Callable[[], PackageInfoManifest]:
    def _assert():
        metadata = sdist_metadata()
        pkginfo_manifest = PackageInfoManifest(metadata.root_folder)
        pkginfo_manifest.read()
        assert_that(pkginfo_manifest.get("Metadata-Version")).is_equal_to("2.4")
        assert_that(pkginfo_manifest.get("Name")).is_equal_to(package_name())
        assert_that(pkginfo_manifest.get("Version")).is_equal_to(package_version())
        return pkginfo_manifest

    return _assert



import textwrap
from enum import Enum
from typing import Optional

from poetry.core._vendor.packaging.version import Version


class MetadataVersions(Enum):
    V1_0 = Version("1.0")
    V1_1 = Version("1.1")
    V1_2 = Version("1.2")
    V2_1 = Version("2.1")
    V2_2 = Version("2.2")
    V2_3 = Version("2.3")
    V2_4 = Version("2.4")

    def __lt__(self, other: "MetadataVersions") -> bool:
        return self.value < other.value

    def __le__(self, other: "MetadataVersions") -> bool:
        return self.value <= other.value

    def __gt__(self, other: "MetadataVersions") -> bool:
        return self.value > other.value

    def __ge__(self, other: "MetadataVersions") -> bool:
        return self.value >= other.value

    def __eq__(self, other: "MetadataVersions") -> bool:
        return self.value == other.value

    def __ne__(self, other: "MetadataVersions") -> bool:
        return self.value != other.value

    def __str__(self) -> str:
        return str(self.value)

    def __repr__(self) -> str:
        return f"MetadataVersions.{self.value}"


class _MetadataField:
    def __init__(
        self,
        name: str,
        version: MetadataVersions,
        required: bool = False,
        multiuse: bool = False,
        multiline: bool = False,
        deprecated_since: Optional[MetadataVersions] = None,
    ):
        self.name = name
        self.version = version
        self.required = required
        self.multiuse = multiuse
        self.multiline = multiline
        self.deprecated_since = deprecated_since

    def format_value(self, value) -> str:
        if self.multiline and isinstance(value, str):
            # Indentation is not only for readability, but required
            # so that the line break is not treated as end of field.
            # The exact indentation does not matter,
            # but it is essential to also indent empty lines.
            from poexy_core.manifest.parser import MULTILINE_PREFIX

            prefix = MULTILINE_PREFIX
            lines = value.splitlines()
            if len(lines) == 1:
                return value
            lines_to_indent = lines[1:]
            value = "\n".join(lines_to_indent)
            value = textwrap.indent(value, prefix, lambda _: True)
            value = f"{lines[0].strip()}\n{value}"
        return value


class MetadataField(Enum):
    # Version 1.0
    MetadataVersion = _MetadataField(
        "Metadata-Version", MetadataVersions.V1_0, required=True
    )
    Name = _MetadataField("Name", MetadataVersions.V1_0, required=True)
    Version = _MetadataField("Version", MetadataVersions.V1_0, required=True)
    Platform = _MetadataField("Platform", MetadataVersions.V1_0, multiuse=True)
    Summary = _MetadataField("Summary", MetadataVersions.V1_0, required=True)
    Description = _MetadataField("Description", MetadataVersions.V1_0, multiline=True)
    Keywords = _MetadataField("Keywords", MetadataVersions.V1_0)
    Author = _MetadataField("Author", MetadataVersions.V1_0, multiline=True)
    AuthorEmail = _MetadataField("Author-email", MetadataVersions.V1_0)
    License = _MetadataField(
        "License",
        MetadataVersions.V1_0,
        multiline=True,
        deprecated_since=MetadataVersions.V2_4,
    )
    HomePage = _MetadataField(
        "Home-Page", MetadataVersions.V1_0, deprecated_since=MetadataVersions.V1_2
    )

    # Version 1.1
    SupportedPlatforms = _MetadataField(
        "Supported-Platforms", MetadataVersions.V1_1, multiuse=True
    )
    Classifier = _MetadataField("Classifier", MetadataVersions.V1_1, multiuse=True)
    DownloadURL = _MetadataField(
        "Download-URL", MetadataVersions.V1_1, deprecated_since=MetadataVersions.V1_2
    )
    Requires = _MetadataField(
        "Requires", MetadataVersions.V1_1, deprecated_since=MetadataVersions.V1_2
    )
    Provides = _MetadataField(
        "Provides", MetadataVersions.V1_1, deprecated_since=MetadataVersions.V1_2
    )
    Obsoletes = _MetadataField(
        "Obsoletes", MetadataVersions.V1_1, deprecated_since=MetadataVersions.V1_2
    )

    # Version 1.2
    Maintainer = _MetadataField("Maintainer", MetadataVersions.V1_2, multiline=True)
    MaintainerEmail = _MetadataField("Maintainer-email", MetadataVersions.V1_2)
    RequiresDist = _MetadataField("Requires-Dist", MetadataVersions.V1_2, multiuse=True)
    RequiresPython = _MetadataField("Requires-Python", MetadataVersions.V1_2)
    RequiresExternal = _MetadataField(
        "Requires-External", MetadataVersions.V1_2, multiuse=True
    )
    ProjectURL = _MetadataField("Project-URL", MetadataVersions.V1_2, multiuse=True)
    ObsoletesDist = _MetadataField(
        "Obsoletes-Dist", MetadataVersions.V1_2, multiuse=True
    )

    # Version 2.1
    DescriptionContentType = _MetadataField(
        "Description-Content-Type", MetadataVersions.V2_1
    )
    ProvidesExtra = _MetadataField(
        "Provides-Extra", MetadataVersions.V2_1, multiuse=True
    )
    ProvidesDist = _MetadataField("Provides-Dist", MetadataVersions.V2_1, multiuse=True)

    # Version 2.2
    Dynamic = _MetadataField("Dynamic", MetadataVersions.V2_2, multiuse=True)

    # Version 2.4
    LicenseExpression = _MetadataField(
        "License-Expression", MetadataVersions.V2_4, multiuse=True
    )
    LicenseFile = _MetadataField("License-File", MetadataVersions.V2_4, multiuse=True)

    @property
    def name(self) -> str:
        return self.value.name

    @property
    def version(self) -> MetadataVersions:
        return self.value.version

    @property
    def required(self) -> bool:
        return self.value.required

    @property
    def multiuse(self) -> bool:
        return self.value.multiuse

    @property
    def multiline(self) -> bool:
        return self.value.multiline

    @property
    def deprecated_since(self) -> Optional[MetadataVersions]:
        return self.value.deprecated_since

    def format_value(self, value) -> str:
        return self.value.format_value(value)



import logging
from contextlib import contextmanager
from typing import Generator

from poetry.core.masonry.metadata import Metadata

from poexy_core.builders.types import FilePathCallback
from poexy_core.packages.format import PackageFormat

logger = logging.getLogger(__name__)


class HookBuilder:
    def __init__(self, name: str):
        self.__name = name

    @contextmanager
    def _hook_build(self) -> Generator[None, None, None]:
        logger.info(f"Building hook: {self.__name}")
        yield
        logger.info(f"Hook {self.__name} built successfully.")

    def build(self):
        raise NotImplementedError("Subclass must implement build method")

    @contextmanager
    def _hook_add_files(self) -> Generator[None, None, None]:
        logger.info(f"Adding files from hook: {self.__name}")
        yield
        logger.info(f"Hook {self.__name} added files successfully.")

    def add_files(self, callback: FilePathCallback):
        raise NotImplementedError("Subclass must implement add_files method")

    @contextmanager
    def _hook_add_metadata(self) -> Generator[None, None, None]:
        logger.info(f"Adding metadata from hook: {self.__name}")
        yield
        logger.info(f"Hook {self.__name} added metadata successfully.")

    def add_metadata(self, metadata: Metadata, _format: PackageFormat):
        raise NotImplementedError("Subclass must implement add_metadata method")



import logging
import shutil
import sys
import zipfile
from contextlib import contextmanager
from enum import Enum
from pathlib import Path
from typing import Any, Generator, Optional, override

from packaging.utils import NormalizedName
from poetry.core.masonry.utils.helpers import distribution_name
from poetry.core.poetry import Poetry

from poexy_core.builders.builder import Builder, PythonTag
from poexy_core.builders.hooks.include_files import IncludeFilesHookBuilder
from poexy_core.builders.hooks.license import LicenseHookBuilder
from poexy_core.builders.hooks.package_files import PackageFilesHookBuilder
from poexy_core.builders.types import FilePathPredicate
from poexy_core.manifest.manifest import MetadataManifest, RecordManifest, WheelManifest
from poexy_core.packages.files import WHEEL_EXTENSIONS
from poexy_core.packages.format import PackageFormat
from poexy_core.pyproject.tables.poexy import Poexy

logger = logging.getLogger(__name__)


class Manifests:
    def __init__(self, path: Path):
        self.metadata: MetadataManifest = MetadataManifest(path)
        self.wheel: WheelManifest = WheelManifest(path)
        self.record: RecordManifest = RecordManifest(path)

    def write(self):
        self.metadata.write()
        self.wheel.write()
        self.record.add(self.metadata.path)
        self.record.add(self.wheel.path)
        self.record.add_self()
        self.record.write()


class WheelDataScheme(str, Enum):
    PURELIB = "purelib"
    PLATLIB = "platlib"
    HEADERS = "headers"
    SCRIPTS = "scripts"
    DATA = "data"


class WheelMetadata:
    def __init__(
        self,
        path: Path,
        name: str,
        version: str,
        python_tag: PythonTag,
        root_folder: Path | None = None,
    ):
        self.path: Path = path
        self.name: str = distribution_name(NormalizedName(name))
        self.version: str = version
        self.python_tag: PythonTag = python_tag
        self.archive_path: Path = self.__archive_path()
        if root_folder is None:
            self.root_folder = self.__root_folder()
        else:
            self.root_folder = root_folder
        self.dist_info_folder: Path = self.__dist_info_folder()
        self.data_purelib_folder: Path = self.__data_folder(WheelDataScheme.PURELIB)
        self.data_platlib_folder: Path = self.__data_folder(WheelDataScheme.PLATLIB)
        self.data_headers_folder: Path = self.__data_folder(WheelDataScheme.HEADERS)
        self.data_scripts_folder: Path = self.__data_folder(WheelDataScheme.SCRIPTS)
        self.data_data_folder: Path = self.__data_folder(WheelDataScheme.DATA)

    def __archive_path(self) -> Path:
        return self.path / Path(f"{self.name}-{self.version}-{self.python_tag}.whl")

    def __root_folder(self) -> Path:
        return self.path / Path(f"{self.name}-{self.version}")

    def __dist_info_folder(self) -> Path:
        return self.root_folder / Path(f"{self.name}-{self.version}.dist-info")

    def __data_folder(self, scheme: WheelDataScheme) -> Path:
        return (
            self.root_folder / Path(f"{self.name}-{self.version}.data") / scheme.value
        )


class WheelBuilder(Builder):
    def __init__(
        self,
        poetry: Poetry,
        poexy: Poexy,
        _format: PackageFormat,
        wheel_directory: Path,
        metadata_directory: Path | None = None,
        config_settings: dict[str, Any] | None = None,
    ):
        if _format != PackageFormat.Wheel:
            raise ValueError(f"Invalid format: {_format}")
        super().__init__(poetry, poexy, _format, wheel_directory, config_settings)
        self._metadata_directory = metadata_directory
        python_tag = PythonTag(impl="py", major=sys.version_info.major)
        self._init_metadata(python_tag)
        self._manifests = Manifests(self._metadata.dist_info_folder)
        self.__archive: Optional[zipfile.ZipFile] = None

        self._hooks.append(
            PackageFilesHookBuilder(
                self.poexy, self.format, self.__filter_file_destination_path()
            )
        )
        self._hooks.append(
            IncludeFilesHookBuilder(
                self.poexy, self.format, self._metadata.data_data_folder
            )
        )
        self._hooks.append(
            LicenseHookBuilder(
                self.poexy,
                self._metadata.dist_info_folder / "licenses",
            )
        )

    def _init_metadata(self, python_tag: PythonTag):
        package_name = self.poetry.package.name
        package_version = str(self.poetry.package.version)
        if self._metadata_directory is not None:
            destination_directory = self._metadata_directory
            root_folder = self._metadata_directory
        else:
            destination_directory = self.temp_destination_directory
            root_folder = None
        self._metadata = WheelMetadata(
            destination_directory,
            package_name,
            package_version,
            python_tag,
            root_folder,
        )

    def __filter_file_destination_path(self) -> FilePathPredicate:
        pure_lib_extensions = ".py"
        platform_lib_extensions = set(WHEEL_EXTENSIONS)

        def predicate(path: Path) -> Optional[Path]:
            if path.suffix in pure_lib_extensions:
                return self._metadata.data_purelib_folder
            if path.suffix in platform_lib_extensions:
                return self._metadata.data_platlib_folder
            if path.suffix not in WHEEL_EXTENSIONS:
                return self._metadata.data_data_folder
            return None

        return predicate

    def _add_files(self):
        def add_file(source: Path, destination: Path):
            logger.info(f"Recording file: {destination.name}")
            self._manifests.record.set(source, destination)
            self._add_files_to_archive(source, destination)

        for hook in self._hooks:
            hook.add_files(add_file)

    def _add_wheel(self):
        self._manifests.wheel.set("Wheel-Version", "1.0")
        self._manifests.wheel.set("Generator", "Poexy")
        self._manifests.wheel.set("Root-Is-Purelib", "true")
        self._manifests.wheel.set("Tag", f"{self._metadata.python_tag}")

    @contextmanager
    def _create_archive(self) -> Generator[None, None, None]:
        logger.info(
            "Creating archive: "
            f"{self._metadata.archive_path.relative_to(self._metadata.path)}"
        )
        self.__archive = zipfile.ZipFile(
            self._metadata.archive_path, "w", zipfile.ZIP_DEFLATED
        )
        yield
        self.__archive.close()
        self.__archive = None

    def _add_files_to_archive(self, source: Path, destination: Path):
        if self.__archive is None:
            raise ValueError("Archive not created")
        relative_path = destination.relative_to(self._metadata.root_folder)
        logger.info(f"Adding: {relative_path}")
        self.__archive.write(source, relative_path)

    def _add_dist_info_files_to_archive(self):
        if self.__archive is None:
            raise ValueError("Archive not created")
        self._add_files_to_archive(
            self._manifests.metadata.path, self._manifests.metadata.path
        )
        self._add_files_to_archive(
            self._manifests.wheel.path, self._manifests.wheel.path
        )
        self._add_files_to_archive(
            self._manifests.record.path, self._manifests.record.path
        )

    def prepare(self) -> Path:
        logger.info("Preparing wheel metadata...")
        logger.info(f"Metadata directory: {self._metadata.dist_info_folder}")
        logger.info("Adding metadata...")
        self._add_metadata(self._manifests.metadata)
        logger.info("Adding wheel...")
        self._add_wheel()
        logger.info("Writing manifests...")
        self._manifests.metadata.write()
        self._manifests.wheel.write()
        logger.info("Wheel metadata prepared successfully.")
        return self._metadata.dist_info_folder

    @override
    @contextmanager
    def build(self) -> Generator[Path, None, None]:
        logger.info("Building wheel...")

        for hook in self._hooks:
            hook.build()

        with self._create_archive():
            logger.info("Adding metadata...")
            self._add_metadata(self._manifests.metadata)
            logger.info("Adding wheel...")
            self._add_wheel()
            logger.info("Adding files...")
            self._add_files()
            logger.info("Writing manifests...")
            self._manifests.write()
            logger.info("Adding dist-info files...")
            self._add_dist_info_files_to_archive()

        logger.info("Built successfully.")

        shutil.copy(self._metadata.archive_path, self.destination_directory)

        yield self._metadata.archive_path



import logging
from pathlib import Path
from typing import Any

from rich.console import Console

from poexy_core.builders.binary import BinaryBuilder
from poexy_core.builders.hooks.binary import BinaryHookBuilder
from poexy_core.builders.sdist import SdistBuilder
from poexy_core.builders.wheel import WheelBuilder
from poexy_core.packages.format import PackageFormat, WheelFormat
from poexy_core.pyproject.toml import PyProjectTOML

console = Console()


class ConsoleHandler(logging.Handler):
    def emit(self, record):
        try:
            msg = self.format(record)
            print(msg)
        except Exception:
            self.handleError(record)


console_handler = ConsoleHandler()
console_handler.setFormatter(logging.Formatter("%(message)s"))

# Configure logging
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

# Configure root logger to show all logs
root_logger = logging.getLogger()
root_logger.setLevel(logging.INFO)

# Also configure poexy_core logger
poexy_logger = logging.getLogger("poexy_core")
poexy_logger.setLevel(logging.INFO)
poexy_logger.addHandler(console_handler)


class PoexyBuilderError(Exception):
    pass


def __get_wheel_builder(
    wheel_directory: str | None = None,
    metadata_directory: str | None = None,
    config_settings: dict[str, Any] | None = None,
) -> WheelBuilder:
    pyproject = PyProjectTOML(path=Path.cwd())
    poexy = pyproject.poexy
    poetry = pyproject.poetry
    if metadata_directory is not None:
        destination_directory = metadata_directory
    elif wheel_directory is not None:
        destination_directory = wheel_directory
    else:
        raise ValueError("wheel_directory and metadata_directory cannot be None")
    metadata_path = None if metadata_directory is None else Path(metadata_directory)
    wheel_builder = WheelBuilder(
        poetry,
        poexy,
        PackageFormat.Wheel,
        wheel_directory=Path(destination_directory),
        metadata_directory=metadata_path,
        config_settings=config_settings,
    )
    if poexy.wheel is None:
        raise PoexyBuilderError(
            "Wheel package is not defined in pyproject.toml at [tool.poexy.wheel]"
        )
    binary_builder = __get_binary_builder(
        wheel_directory=wheel_directory,
        metadata_directory=metadata_directory,
        config_settings=config_settings,
    )
    if poexy.wheel.format is not None:
        wheel_format = list(poexy.wheel.format)
        if len(wheel_format) > 1 and WheelFormat.Binary in wheel_format:
            wheel_builder.add_hook(BinaryHookBuilder(binary_builder))
            return wheel_builder
        if len(wheel_format) == 1 and WheelFormat.Binary not in wheel_format:
            return wheel_builder
        if len(wheel_format) == 1 and WheelFormat.Binary in wheel_format:
            return binary_builder
        raise PoexyBuilderError("Wheel package format is not valid")
    raise PoexyBuilderError(
        "Wheel package format is not defined in pyproject.toml at " "[tool.poexy.wheel]"
    )


def __get_binary_builder(
    wheel_directory: str | None = None,
    metadata_directory: str | None = None,
    config_settings: dict[str, Any] | None = None,
) -> BinaryBuilder:
    pyproject = PyProjectTOML(path=Path.cwd())
    poexy = pyproject.poexy
    poetry = pyproject.poetry
    if metadata_directory is not None:
        destination_directory = metadata_directory
    elif wheel_directory is not None:
        destination_directory = wheel_directory
    else:
        raise ValueError("wheel_directory and metadata_directory cannot be None")
    metadata_path = None if metadata_directory is None else Path(metadata_directory)
    builder = BinaryBuilder(
        poetry,
        poexy,
        PackageFormat.Binary,
        wheel_directory=Path(destination_directory),
        metadata_directory=metadata_path,
        config_settings=config_settings,
    )
    return builder


def __get_sdist_builder(
    sdist_directory: str,
    config_settings: dict[str, Any] | None = None,
) -> SdistBuilder:
    pyproject = PyProjectTOML(path=Path.cwd())
    poexy = pyproject.poexy
    poetry = pyproject.poetry
    builder = SdistBuilder(
        poetry,
        poexy,
        PackageFormat.Source,
        sdist_directory=Path(sdist_directory),
        config_settings=config_settings,
    )
    return builder


def get_requires_for_build_wheel(
    config_settings: dict[str, Any] | None = None,
) -> list[str]:
    """
    Returns an additional list of requirements for building, as PEP508 strings,
    above and beyond those specified in the pyproject.toml file.

    This implementation is optional. At the moment it only returns an empty list,
    which would be the same as if not define. So this is just for completeness
    for future implementation.
    """

    logger.info("=== Get requires for build wheel called ===")
    logger.info(f"config_settings: {config_settings}")

    try:
        pyproject = PyProjectTOML(path=Path().resolve())
        logger.info("Pyproject loaded successfully")
        required_packages = pyproject.build_system.requires
        logger.info(f"Required packages: {required_packages}")
        return required_packages
    except Exception as e:
        logger.error(f"Error in get_requires_for_build_wheel: {e}")
        raise


# For now, we require all dependencies to build either a wheel or an sdist.
get_requires_for_build_sdist = get_requires_for_build_wheel


def prepare_metadata_for_build_wheel(
    metadata_directory: str, config_settings: dict[str, Any] | None = None
) -> str:
    logger.info("=== Prepare metadata for build wheel called ===")
    logger.info(f"metadata_directory: {metadata_directory}")
    logger.info(f"config_settings: {config_settings}")

    try:
        builder = __get_wheel_builder(
            metadata_directory=metadata_directory, config_settings=config_settings
        )
        logger.info("Builder created successfully")
        dist_info = builder.prepare()
        logger.info(f"Metadata prepared, dist_info: {dist_info}")
        logger.info(f"Returning dist_info.name: {dist_info.name}")
        return dist_info.name
    except Exception as e:
        logger.error(f"Error in prepare_metadata_for_build_wheel: {e}")
        raise


def build_wheel(
    wheel_directory: str,
    config_settings: dict[str, Any] | None = None,
    metadata_directory: str | None = None,
) -> str:
    """Builds a wheel, places it in wheel_directory"""
    logger.info("=== Build wheel called ===")
    logger.info(f"wheel_directory: {wheel_directory}")
    logger.info(f"config_settings: {config_settings}")
    logger.info(f"metadata_directory: {metadata_directory}")

    try:
        builder = __get_wheel_builder(
            wheel_directory=wheel_directory, config_settings=config_settings
        )
        logger.info("Builder created successfully")
        with builder.build() as file_path:
            logger.info(f"Wheel built successfully, file_path: {file_path}")
            return file_path.name
    except Exception as e:
        logger.error(f"Error in build_wheel: {e}")
        raise


def build_sdist(
    sdist_directory: str, config_settings: dict[str, Any] | None = None
) -> str:
    """Builds an sdist, places it in sdist_directory"""
    logger.info("=== Build sdist called ===")
    logger.info(f"sdist_directory: {sdist_directory}")
    logger.info(f"config_settings: {config_settings}")

    try:
        builder = __get_sdist_builder(
            sdist_directory=sdist_directory, config_settings=config_settings
        )
        logger.info("Builder created successfully")
        with builder.build() as file_path:
            logger.info(f"Sdist built successfully, file_path: {file_path}")
            return file_path.name
    except Exception as e:
        logger.error(f"Error in build_sdist: {e}")
        raise


get_requires_for_build_editable = get_requires_for_build_wheel
# prepare_metadata_for_build_editable = prepare_metadata_for_build_wheel



from pathlib import Path
from typing import override

from poetry.core.masonry.metadata import Metadata

from poexy_core.builders.hooks.hook import HookBuilder
from poexy_core.builders.types import FilePathCallback
from poexy_core.packages.format import PackageFormat
from poexy_core.pyproject.tables.poexy import Poexy


class LicenseHookBuilder(HookBuilder):
    def __init__(self, poexy: Poexy, destination_path: Path):
        super().__init__("license")
        self.__poexy = poexy
        self.__destination_path = destination_path
        self.__files = []

    @override
    def build(self):
        with self._hook_build():
            if self.__poexy.license is None:
                return
            if self.__poexy.license.files is None:
                return
            if len(self.__poexy.license.files) == 0:
                return
            files = self.__poexy.license.resolve_files()
            if files is None:
                return
            for file in files:
                source = Path(file).resolve()
                destination = self.__destination_path / Path(file).name
                self.__files.append((source, destination))

    @override
    def add_files(self, callback: FilePathCallback):
        with self._hook_add_files():
            for source, destination in self.__files:
                callback(source, destination)

    @override
    def add_metadata(self, metadata: Metadata, _format: PackageFormat):
        with self._hook_add_metadata():
            if self.__poexy.license is None:
                return
            self.__poexy.license.transform_poetry_metadata(metadata, _format)



from typing import Any, Dict

from poetry.core.masonry.utils.helpers import distribution_name

from poexy_core.pyproject.exceptions import PyProjectError


class ModulePackage:
    @staticmethod
    def __transform_config(config: Dict[str, Any]) -> Dict[str, Any]:
        package_config = config.pop("package", None)
        if package_config is None:
            raise PyProjectError("[tool.poexy.package] section not found")
        assert isinstance(package_config, dict)
        if len(package_config.keys()) == 0:
            raise PyProjectError("[tool.poexy.package] section not found")
        if len(package_config.keys()) > 1:
            raise PyProjectError(
                "[tool.poexy.package] section must contain only one package definition"
            )
        package_name, package_config = package_config.popitem()
        package_config = {"name": package_name, **package_config}
        config["package"] = package_config
        return config

    @staticmethod
    def from_project_config(config: Dict[str, Any]) -> Dict[str, Any]:
        project = config.get("project", None)
        assert isinstance(project, dict)
        project_name = project.get("name", None)
        if project_name is None:
            raise PyProjectError("[project] section must contain a name")
        project_name = distribution_name(project_name)
        config["package"] = {
            project_name: {
                "source": project_name,
            }
        }
        return ModulePackage.__transform_config(config)

    @staticmethod
    def from_poexy_config(config: Dict[str, Any]) -> Dict[str, Any]:
        return ModulePackage.__transform_config(config)


class BinaryPackage:
    @staticmethod
    def __transform_config(config: Dict[str, Any]) -> Dict[str, Any]:
        binary_config = config.get("binary", None)
        if binary_config is None:
            raise PyProjectError("[tool.poexy.binary] section not found")
        assert isinstance(binary_config, dict)
        executable_name = binary_config.get("name", None)
        if executable_name is None:
            package_name = config.get("name", None)
            if package_name is None:
                raise PyProjectError("[project] section must contain a name")
            executable_name = package_name.replace("_", "-")
            binary_config["name"] = executable_name
        else:
            binary_config["name"] = executable_name.replace("_", "-")
        return config

    @staticmethod
    def from_project_config(config: Dict[str, Any]) -> Dict[str, Any]:
        project = config.get("project", None)
        assert isinstance(project, dict)
        project_name = project.get("name", None)
        if project_name is None:
            raise PyProjectError("[project] section must contain a name")
        config["binary"] = {"name": project_name}
        return BinaryPackage.__transform_config(config)

    @staticmethod
    def from_poexy_config(config: Dict[str, Any]) -> Dict[str, Any]:
        binary_config = config.get("binary", None)
        if binary_config is None:
            binary_config = {"name": None}
            config["binary"] = binary_config
        assert isinstance(binary_config, dict)
        if len(binary_config.keys()) == 0:
            raise PyProjectError("[tool.poexy.binary] section not found")
        if len(binary_config.keys()) > 1:
            raise PyProjectError(
                "[tool.poexy.binary] section must contain only one binary definition"
            )
        return BinaryPackage.__transform_config(config)



# pylint: disable=wildcard-import,unused-wildcard-import
# flake8: noqa: F403,F401

from conftests.assert_builds import *
from conftests.assert_manifests import *
from conftests.logger import *
from conftests.metadata import *
from conftests.paths import *
from conftests.pip import *
from conftests.project import *



from pathlib import Path

import pytest
from assertpy import assert_that

from poexy_core.metadata.fields import MetadataField

# pylint: disable=redefined-outer-name


@pytest.fixture()
def project_path(sample_project):
    return sample_project("license-file")


def test_wheel(
    project,
    project_path,
    assert_wheel_build,
    wheel_data_purelib_folder,
    wheel_dist_info_folder,
    default_python_tag,
    dist_package_name,
    site_packages_path,
    assert_metadata_manifest,
):
    with project(project_path):
        assert_zip_file = assert_wheel_build(project_path)
        assert_zip_file(
            [
                wheel_data_purelib_folder(default_python_tag)
                / dist_package_name()
                / "__init__.py",
                wheel_dist_info_folder(default_python_tag) / "licenses" / "LICENSE",
            ],
            strict=True,
        )
        purelib_path = site_packages_path / dist_package_name() / "__init__.py"
        assert_that(purelib_path.exists()).is_true()
        metadata_manifest = assert_metadata_manifest(default_python_tag)
        assert_that(metadata_manifest.get(MetadataField.LicenseFile)).is_equal_to(
            "licences/LICENSE"
        )


def test_sdist(project, project_path, assert_sdist_build, assert_pkginfo_manifest):
    with project(project_path):
        assert_tar_file = assert_sdist_build(project_path)
        assert_tar_file(
            [
                Path("src") / "__init__.py",
                Path("LICENSE"),
            ],
            strict=True,
        )
        pkginfo_manifest = assert_pkginfo_manifest()
        assert_that(pkginfo_manifest.get(MetadataField.LicenseFile)).is_equal_to(
            "LICENSE"
        )



from pathlib import Path
from typing import Any, List, Optional, override

from pydantic import BaseModel, Field, model_validator

from poexy_core.packages.files import PackageFile, PackageFiles, ResolvedPackageFiles
from poexy_core.packages.format import PackageFormat
from poexy_core.packages.package import (
    BinaryPackage,
    ModulePackage,
    SdistPackage,
    WheelPackage,
)
from poexy_core.pyproject.tables.license import License
from poexy_core.pyproject.tables.readme import Readme

# pylint: disable=no-member,attribute-defined-outside-init


class Poexy(BaseModel):
    package: ModulePackage = Field(
        description="Package to include in the built archive"
    )
    wheel: Optional[WheelPackage] = Field(
        description="Wheel package configuration", default=None
    )
    sdist: Optional[SdistPackage] = Field(
        description="Sdist package configuration", default=None
    )
    binary: Optional[BinaryPackage] = Field(
        description="Binary package configuration", default=None
    )
    readme: Optional[Readme] = Field(description="Readme configuration", default=None)
    license: Optional[License] = Field(
        description="License configuration", default=None
    )

    @override
    def model_post_init(self, __context: Any, /) -> None:
        self.__resolved_package_files: Optional[PackageFiles] = None
        self.__resolved_inclusions_files: Optional[ResolvedPackageFiles] = None

    @model_validator(mode="after")
    def validate_model(self) -> "Poexy":
        if self.wheel is None:
            self.wheel = WheelPackage()
        return self

    def resolve_package_files(self, _format: PackageFormat) -> PackageFiles:
        if self.__resolved_package_files is None:
            base_path = Path(self.package.name)
            resolved_files = self.package.resolve(_format, base_path)
            if resolved_files is None:
                raise ValueError("No package files found")
            self.__resolved_package_files = resolved_files
        return self.__resolved_package_files

    def resolve_inclusions(self, _format: PackageFormat) -> ResolvedPackageFiles:
        if self.__resolved_inclusions_files is None:
            includes: List[PackageFile] = []
            excludes: List[PackageFile] = []
            base_path = Path(self.package.name)
            if self.package.includes is not None:
                resolved_files = self.package.resolve_includes(_format, base_path)
                if resolved_files is not None:
                    includes.extend(resolved_files)
            if self.package.excludes is not None:
                resolved_files = self.package.resolve_excludes(_format, base_path)
                if resolved_files is not None:
                    excludes.extend(resolved_files)
            if self.wheel is not None and _format in PackageFormat.Wheel:
                if self.wheel.includes is not None:
                    resolved_files = self.wheel.resolve_includes(_format, base_path)
                    if resolved_files is not None:
                        includes.extend(resolved_files)
                if self.wheel.excludes is not None:
                    resolved_files = self.wheel.resolve_excludes(_format, base_path)
                    if resolved_files is not None:
                        excludes.extend(resolved_files)
            if self.sdist is not None and _format in PackageFormat.Source:
                if self.sdist.includes is not None:
                    resolved_files = self.sdist.resolve_includes(_format, base_path)
                    if resolved_files is not None:
                        includes.extend(resolved_files)
                if self.sdist.excludes is not None:
                    resolved_files = self.sdist.resolve_excludes(_format, base_path)
                    if resolved_files is not None:
                        excludes.extend(resolved_files)
            if self.binary is not None and _format in PackageFormat.Binary:
                if self.binary.includes is not None:
                    resolved_files = self.binary.resolve_includes(_format, base_path)
                    if resolved_files is not None:
                        includes.extend(resolved_files)
                if self.binary.excludes is not None:
                    resolved_files = self.binary.resolve_excludes(_format, base_path)
                    if resolved_files is not None:
                        excludes.extend(resolved_files)
            self.__resolved_inclusions_files = ResolvedPackageFiles(
                includes=set(includes), excludes=set(excludes)
            )
        return self.__resolved_inclusions_files



import subprocess
from typing import Callable, List

Printer = Callable[[str], None]


def run(cmd: List[str], printer: Printer, **kwargs) -> int:
    arguments = {
        "stdout": subprocess.PIPE,
        "stderr": subprocess.STDOUT,
        "shell": False,
        "text": True,
        "encoding": "utf-8",
    }

    arguments.update(kwargs)

    process = subprocess.Popen(cmd, **arguments)

    while True:
        if process.stdout is None:
            continue
        output = process.stdout.readline()
        if output and output != "":
            printer(output.strip())
            continue
        if process.poll() is not None:
            break

    exit_code = process.poll()

    if exit_code is None:
        raise RuntimeError("Process terminated unexpectedly")

    return exit_code



import logging

import pytest

# pylint: disable=redefined-outer-name


@pytest.fixture(scope="session")
def root_logger():
    logger = logging.getLogger()
    logger.setLevel(logging.INFO)
    logger.handlers.clear()
    logger.addHandler(logging.StreamHandler())
    return logger


@pytest.fixture(autouse=True)
def set_logger_level(caplog):
    caplog.at_level(logging.INFO)


@pytest.fixture(scope="session")
def log_info(root_logger):
    def _log_info(message):
        root_logger.info(message)

    return _log_info


@pytest.fixture(scope="session")
def log_info_section(log_info):
    def _log_info_section(message):
        pad = "=" * 20
        log_info(f"\n{pad} {message} {pad}\n")

    return _log_info_section



from pathlib import Path

import pytest
from assertpy import assert_that

from poexy_core.metadata.fields import MetadataField

# pylint: disable=redefined-outer-name


@pytest.fixture()
def project_path(sample_project):
    return sample_project("readme-markdown")


def test_wheel(
    project,
    project_path,
    assert_wheel_build,
    wheel_data_purelib_folder,
    default_python_tag,
    dist_package_name,
    site_packages_path,
):
    with project(project_path):
        assert_zip_file = assert_wheel_build(project_path)
        assert_zip_file(
            [
                wheel_data_purelib_folder(default_python_tag)
                / dist_package_name()
                / "__init__.py"
            ],
            strict=True,
        )
        purelib_path = site_packages_path / dist_package_name() / "__init__.py"
        assert_that(purelib_path.exists()).is_true()


def test_sdist(project, project_path, assert_sdist_build, assert_pkginfo_manifest):
    with project(project_path):
        assert_tar_file = assert_sdist_build(project_path)
        assert_tar_file(
            [
                Path("src") / "__init__.py",
                Path("README.md"),
            ],
            strict=True,
        )
        pkginfo_manifest = assert_pkginfo_manifest()
        assert_that(
            pkginfo_manifest.get(MetadataField.DescriptionContentType)
        ).is_equal_to("text/markdown")
        assert_that(pkginfo_manifest.get(MetadataField.Description)).is_equal_to(
            "# I'm super informative\n\nAnd also multilines"
        )



from pathlib import Path
from typing import List, Optional, Set, override

from pydantic import BaseModel, Field, model_validator

from poexy_core.packages.files import FilePattern, PackageFiles, ResolvePackageFiles
from poexy_core.packages.format import (
    DEFAULT_FORMATS,
    DEFAULT_WHEEL_FORMATS,
    PackageFormat,
    WheelFormat,
)
from poexy_core.packages.inclusions import Excludes, Includes
from poexy_core.packages.validators import validate_path

# pylint: disable=no-member


class BasePackage(BaseModel):
    includes: Optional[Includes] = Field(
        description="List of includes in package", default=None
    )
    excludes: Optional[Excludes] = Field(
        description="List of excludes in package", default=None
    )
    _is_format_allowed: bool = False
    _default_formats: List[PackageFormat]

    @model_validator(mode="after")
    def validate_package(self) -> "BasePackage":
        if self.includes is not None:
            for include in self.includes.root:
                if include.formats is not None and not self._is_format_allowed:
                    raise ValueError("formats are not allowed in includes")
                if include.formats is None:
                    include.formats = self._default_formats
        if self.excludes is not None:
            for exclude in self.excludes.root:
                if exclude.formats is not None and not self._is_format_allowed:
                    raise ValueError("formats are not allowed in excludes")
                if exclude.formats is None:
                    exclude.formats = self._default_formats
        return self

    def resolve_includes(
        self, _format: PackageFormat, base_path: Path
    ) -> Optional[PackageFiles]:
        if self.includes is None:
            return None
        return self.includes.resolve(_format, base_path)

    def resolve_excludes(
        self, _format: PackageFormat, base_path: Path
    ) -> Optional[PackageFiles]:
        if self.excludes is None:
            return None
        return self.excludes.resolve(_format, base_path)


class ModulePackage(BasePackage, ResolvePackageFiles):
    _is_format_allowed: bool = True
    _default_formats: List[PackageFormat] = DEFAULT_FORMATS

    name: str = Field(description="Name of the package")
    source: Optional[Path] = Field(
        description="Source path if it differs from the name of the package",
        default=None,
    )

    @model_validator(mode="after")
    def validate_module_package(self) -> "ModulePackage":
        if self.includes is not None:
            for include in self.includes.root:
                if include.formats is None:
                    include.formats = self._default_formats
        if self.excludes is not None:
            for exclude in self.excludes.root:
                if exclude.formats is None:
                    exclude.formats = self._default_formats
        if self.source is not None:
            self.source = validate_path("source", self.source)
        else:
            self.source = Path(self.name)
        return self

    @override
    def resolve(
        self, _format: PackageFormat, base_path: Path
    ) -> Optional[PackageFiles]:
        if self.source is None:
            raise ValueError("Source is required")
        file_pattern = FilePattern(glob_pattern=Path("**/*"), path=self.source)
        if _format == PackageFormat.Source:
            destination_path = self.source
        else:
            destination_path = Path(self.name)
        resolved = file_pattern.resolve(destination_path)
        return resolved


class WheelPackage(BasePackage):
    _default_formats: List[PackageFormat] = [PackageFormat.Wheel]

    format: Set[WheelFormat] = Field(
        description="Formats of the wheel",
        default=set(DEFAULT_WHEEL_FORMATS),
        min_length=1,
        max_length=2,
    )


class SdistPackage(BasePackage):
    _default_formats: List[PackageFormat] = [PackageFormat.Source]


class BinaryPackage(BasePackage):
    _default_formats: List[PackageFormat] = [PackageFormat.Binary]

    name: Optional[str] = Field(
        default=None,
        min_length=3,
        pattern=r"^[a-zA-Z][a-zA-Z0-9_-]+$",
        description="Name of the executable",
    )

    entry_point: Optional[str] = Field(
        default=None,
        pattern=r"^([a-zA-Z][a-zA-Z0-9_.]+|\.)(:[a-zA-Z0-9_]+)?$",
        description="Entry point of the package",
    )



from poetry.core.masonry.metadata import Metadata as PoetryMetadata

from poexy_core.manifest.manifest import Manifest, ManifestValue
from poexy_core.metadata.fields import MetadataField, MetadataVersions


class MetadataManifestBuilder:
    def __init__(self, manifest: Manifest, version: MetadataVersions):
        self.__manifest = manifest
        self.__version = version
        self.__manifest.set(MetadataField.MetadataVersion.name, str(version))

    def get(self, field: MetadataField) -> ManifestValue:
        return self.__manifest.get(field)

    def set(self, field: MetadataField, value: ManifestValue):
        if field.version > self.__version:
            raise ValueError(
                f"Field {field.name} is not supported in version {self.__version}"
            )
        if field.deprecated_since and field.deprecated_since <= self.__version:
            raise ValueError(
                f"Field {field.name} is deprecated since {field.deprecated_since}"
            )
        self.__manifest.set(field.name, field.format_value(value))

    def delete(self, field: MetadataField):
        self.__manifest.delete(field)

    def validate(self):
        required_fields = [field for field in MetadataField if field.required]
        for field in required_fields:
            if not self.__manifest.get(field.name):
                raise ValueError(f"Field {field.name} is required")

    @staticmethod
    def from_poetry(builder: "MetadataManifestBuilder", metadata: PoetryMetadata):
        if metadata.name is None:
            raise ValueError("Metadata.name is required")
        builder.set(MetadataField.Name, metadata.name)
        builder.set(MetadataField.Version, metadata.version)

        # 1.0
        if metadata.platforms:
            builder.set(MetadataField.Platform, metadata.platforms)

        if metadata.supported_platforms:
            builder.set(MetadataField.SupportedPlatforms, metadata.supported_platforms)

        if metadata.summary:
            builder.set(MetadataField.Summary, metadata.summary)

        if metadata.description:
            builder.set(MetadataField.Description, metadata.description)

        if metadata.keywords:
            builder.set(MetadataField.Keywords, metadata.keywords)

        if metadata.home_page:
            builder.set(MetadataField.HomePage, metadata.home_page)

        if metadata.download_url:
            builder.set(MetadataField.DownloadURL, metadata.download_url)

        if metadata.author:
            builder.set(MetadataField.Author, metadata.author)

        if metadata.author_email:
            builder.set(MetadataField.AuthorEmail, metadata.author_email)

        if metadata.license:
            builder.set(MetadataField.License, metadata.license)
        elif hasattr(metadata, "license_expression"):
            builder.set(
                MetadataField.LicenseExpression, getattr(metadata, "license_expression")
            )
        elif hasattr(metadata, "license_files"):
            builder.set(MetadataField.LicenseFile, getattr(metadata, "license_files"))

        # 1.1
        if metadata.classifiers:
            builder.set(MetadataField.Classifier, metadata.classifiers)

        if metadata.requires:
            builder.set(MetadataField.Requires, metadata.requires)

        if metadata.provides:
            builder.set(MetadataField.Provides, metadata.provides)

        if metadata.obsoletes:
            builder.set(MetadataField.Obsoletes, metadata.obsoletes)

        # 1.2
        if metadata.maintainer:
            builder.set(MetadataField.Maintainer, metadata.maintainer)

        if metadata.maintainer_email:
            builder.set(MetadataField.MaintainerEmail, metadata.maintainer_email)

        if metadata.requires_python:
            builder.set(MetadataField.RequiresPython, metadata.requires_python)

        if metadata.requires_external:
            builder.set(MetadataField.RequiresExternal, metadata.requires_external)

        if metadata.requires_dist:
            builder.set(MetadataField.RequiresDist, sorted(metadata.requires_dist))

        if metadata.provides_dist:
            builder.set(MetadataField.ProvidesDist, metadata.provides_dist)

        if metadata.obsoletes_dist:
            builder.set(MetadataField.ObsoletesDist, metadata.obsoletes_dist)

        if metadata.project_urls:
            builder.set(
                MetadataField.ProjectURL,
                sorted(metadata.project_urls, key=lambda u: u[0]),
            )

        # 2.1
        if metadata.description_content_type:
            builder.set(
                MetadataField.DescriptionContentType, metadata.description_content_type
            )

        if metadata.provides_extra:
            builder.set(MetadataField.ProvidesExtra, sorted(metadata.provides_extra))



from pathlib import Path

import pytest
from assertpy import assert_that

from poexy_core.packages.files import FORBIDDEN_DIRS


@pytest.fixture()
def project_path(sample_project):
    return sample_project("includes")


def test_wheel(
    project,
    project_path,
    install_path,
    site_packages_path,
    dist_package_name,
    default_python_tag,
    wheel_data_purelib_folder,
    assert_wheel_build,
):
    with project(project_path):
        assert_zip_file = assert_wheel_build(project_path)
        assert_zip_file(
            [
                wheel_data_purelib_folder(default_python_tag)
                / dist_package_name()
                / "__init__.py"
            ]
        )

        purelib_path = site_packages_path / "includes"
        data_path = install_path / "share" / "includes"

        assert_that(purelib_path.exists()).is_true()
        assert_that(data_path.exists()).is_true()

        purelib_path = purelib_path.relative_to(install_path)
        data_path = data_path.relative_to(install_path)
        purelib_glob_pattern = f"{purelib_path}/**/*"
        data_glob_pattern = f"{data_path}/**/*"

        for file in install_path.rglob(purelib_glob_pattern):
            if not file.is_file():
                continue
            if any(part in FORBIDDEN_DIRS for part in file.parts):
                continue
            assert_that(file.suffix).is_equal_to(".py")

        for file in install_path.rglob(data_glob_pattern):
            if not file.is_file():
                continue
            if any(part in FORBIDDEN_DIRS for part in file.parts):
                continue
            assert_that(file.suffix).is_equal_to(".md")


def test_sdist(project, project_path, assert_sdist_build):
    with project(project_path):
        assert_tar_file = assert_sdist_build(project_path)
        assert_tar_file(
            [
                Path("module.py"),
                Path("src") / "__init__.py",
                Path("docs") / "test.md",
                Path("docs") / "test.txt",
                Path("docs") / "subdir" / "test.txt",
                Path("docs") / "subdir" / "test.rst",
                Path("docs") / "subdir" / "subsubdir" / "test.txt",
                Path("docs") / "subdir" / "subsubdir" / "test.md",
            ],
            strict=True,
        )



import ast
import logging
from enum import Enum
from pathlib import Path
from typing import Optional

from poetry.core.poetry import Poetry

from poexy_core.packages.format import WheelFormat
from poexy_core.pyproject.tables.poexy import Poexy
from poexy_core.utils import subprocess_rt

logger = logging.getLogger(__name__)


class PyinstallerBuilderError(Exception):
    pass


class BuildType(Enum):
    OneFile = "onefile"
    OneDir = "onedir"


class PyinstallerBuilder:
    def __init__(self, poetry: Poetry, poexy: Poexy):
        self.__project_path = poetry.pyproject_path.parent
        if poexy.package.source is None:
            raise PyinstallerBuilderError(
                "Package source is not defined in pyproject.toml at "
                "[tool.poexy.package.${package_name}.source]"
            )
        self.__package_source_path = poexy.package.source
        wheel = poexy.wheel
        if wheel is not None and WheelFormat.Binary not in wheel.format:
            raise PyinstallerBuilderError(
                "Cannot build binary executable for wheel package as specified in "
                "[tool.poexy.wheel.format]"
            )
        binary = poexy.binary
        if binary is None:
            raise PyinstallerBuilderError(
                "Binary package is not defined in pyproject.toml at [tool.poexy.binary]"
            )
        if binary.name is None:
            raise PyinstallerBuilderError(
                "Binary package name is not defined in pyproject.toml at "
                "[tool.poexy.binary.name]"
            )
        self.__executable_name = binary.name
        self.__entry_point = binary.entry_point

    @property
    def executable_name(self) -> str:
        return self.__executable_name

    @staticmethod
    def __has_main_block(path: Path) -> bool:
        with open(path, "r", encoding="utf-8") as f:
            tree = ast.parse(f.read(), filename=path)
        for node in ast.walk(tree):
            if isinstance(node, ast.If):
                test = node.test
                if (
                    isinstance(test, ast.Compare)
                    and isinstance(test.left, ast.Name)
                    and test.left.id == "__name__"
                    and len(test.ops) == 1
                    and isinstance(test.ops[0], ast.Eq)
                    and len(test.comparators) == 1
                    and isinstance(test.comparators[0], ast.Constant)
                    and test.comparators[0].value == "__main__"
                ):
                    return True
        return False

    @staticmethod
    def __find_entry_point(base_path: Path) -> str:
        for file in base_path.rglob("**/*.py"):
            if not file.is_file():
                continue
            if PyinstallerBuilder.__has_main_block(file):
                relative_path = file.relative_to(base_path)
                file_name = relative_path.stem
                path = str(file.parent).replace("/", ".")
                return f"{path}:{file_name}"
        raise PyinstallerBuilderError("No entry point found")

    def build(
        self,
        build_type: Optional[BuildType] = BuildType.OneFile,
        build_path: Optional[Path] = None,
        dist_path: Optional[Path] = None,
        strip: Optional[bool] = True,
        clean: Optional[bool] = True,
    ):
        if self.__entry_point is None:
            self.__entry_point = PyinstallerBuilder.__find_entry_point(
                self.__package_source_path
            )
        entry_point = self.__entry_point.replace(".", "/").replace(":", "/")
        path_to_entry_point = str(self.__project_path / f"{entry_point}.py")
        spec_path = str(build_path or self.__project_path / "build" / "spec")
        if dist_path is None:
            dist_path = self.__project_path / "dist"
        work_path = str(build_path or self.__project_path / "build" / "temp")

        if build_type == BuildType.OneFile:
            build_type_argument = "--onefile"
        elif build_type == BuildType.OneDir:
            build_type_argument = "--onedir"
        else:
            raise PyinstallerBuilderError(f"Invalid build type: {build_type}")

        logger.info("Building executable:")
        logger.info(f"  - Path to entry point: {path_to_entry_point}")
        logger.info(f"  - Spec path: {spec_path}")
        logger.info(f"  - Dist path: {dist_path}")
        logger.info(f"  - Work path: {work_path}")
        logger.info(f"  - Build type: {build_type}")
        logger.info(f"  - Strip: {strip}")
        logger.info(f"  - Clean: {clean}")

        command = [
            "poetry",
            "run",
            "pyinstaller",
            path_to_entry_point,
            build_type_argument,
            "--name",
            self.__executable_name,
            "--specpath",
            spec_path,
            "--distpath",
            dist_path,
            "--workpath",
            work_path,
            "--strip" if strip else None,
            "--clean" if clean else None,
        ]

        exit_code = subprocess_rt.run(
            command, printer=logger.info, cwd=self.__project_path
        )

        if exit_code != 0:
            raise PyinstallerBuilderError("Failed to build executable")

        logger.info("Executable built successfully.")



from pathlib import Path


def validate_path(field: str, value: Path) -> Path:
    if "*" in str(value):
        raise ValueError(f"{field} path '{value}' is a glob pattern")
    if value.is_absolute():
        raise ValueError(f"{field} path '{value}' is absolute")
    if not value.exists():
        raise ValueError(f"{field} path '{value}' does not exist")
    return value


def validate_destination(field: str, value: Path) -> Path:
    if "*" in str(value):
        raise ValueError(f"{field} path '{value}' is a glob pattern")
    if value.is_absolute():
        raise ValueError(f"{field} path '{value}' is absolute")
    return value



import pytest

# pylint: disable=redefined-outer-name


@pytest.fixture()
def project_path(sample_project):
    return sample_project("license-deprecated-text")


def test_wheel(project, project_path, assert_wheel_build):
    with project(project_path):
        with pytest.raises(ValueError, match="Field License is deprecated since 2.4"):
            assert_wheel_build(project_path)


def test_sdist(project, project_path, assert_sdist_build):
    with project(project_path):
        with pytest.raises(ValueError, match="Field License is deprecated since 2.4"):
            assert_sdist_build(project_path)



from typing import Callable, List, Tuple, Union

from poexy_core.metadata.fields import MetadataField

ManifestKey = Union[str, MetadataField]
ManifestValue = Union[str, List[str], Tuple[str, ...]]
ManifestStorage = List[Tuple[str, ManifestValue]]
ManifestStorageOperation = Callable[[ManifestStorage], None]



import logging
from pathlib import Path
from typing import override

from poetry.core.masonry.metadata import Metadata

from poexy_core.builders.hooks.hook import HookBuilder
from poexy_core.builders.types import FilePathCallback
from poexy_core.packages.format import PackageFormat
from poexy_core.pyproject.tables.poexy import Poexy

logger = logging.getLogger(__name__)


class IncludeFilesHookBuilder(HookBuilder):
    def __init__(
        self,
        poexy: Poexy,
        _format: PackageFormat,
        destination: Path,
    ):
        super().__init__("include_files")
        self.__poexy = poexy
        self.__format = _format
        self.__destination = destination
        self.__files = []

    @override
    def build(self):
        with self._hook_build():
            logger.info("Resolving include files...")
            resolved = self.__poexy.resolve_inclusions(self.__format)
            exclusions = [file.source for file in resolved.excludes]

            logger.info(f"Resolved {len(resolved.includes)} files to include.")
            logger.info(f"Resolved {len(resolved.excludes)} files to exclude.")

            count = 0

            for file in resolved.includes:
                if file.source in exclusions:
                    logger.info(f"Excluding file: {file.source}")
                    continue
                logger.info(f"Including file: {file.source}")
                self.__files.append(
                    (file.source, self.__destination / file.destination)
                )
                count += 1

            logger.info(f"{count} files ready to be packaged.")

    @override
    def add_files(self, callback: FilePathCallback):
        with self._hook_add_files():
            for source, destination in self.__files:
                callback(source, destination)

    @override
    def add_metadata(self, metadata: Metadata, _format: PackageFormat):
        pass



#!/usr/bin/env python3
"""
Python Implementation Detection Script

This script detects the current Python implementation and generates
the appropriate wheel tag according to PEP 425 standards.
"""

import platform
import sys
from typing import Optional


def get_python_implementation_tag() -> str:
    """
    Detect the Python implementation and return the appropriate tag.

    Returns:
        str: The implementation tag (py, cp, ip, pp, jy, etc.)
    """
    impl = sys.implementation

    # Standard implementation tags according to PEP 425
    impl_tags = {
        "cpython": "cp",
        "pypy": "pp",
        "ironpython": "ip",
        "jython": "jy",
        "micropython": "mp",
        "graalpython": "gp",
        "stackless": "sl",
    }

    # Get the implementation name
    impl_name = impl.name.lower()

    # Return the corresponding tag, default to 'py' for unknown implementations
    return impl_tags.get(impl_name, "py")


def get_python_version_info() -> tuple[int, int]:
    """
    Get the current Python version information.

    Returns:
        tuple[int, int]: (major, minor) version numbers
    """
    return sys.version_info.major, sys.version_info.minor


def get_abi_tag() -> str:
    """
    Get the ABI tag for the current Python implementation.

    Returns:
        str: The ABI tag
    """
    impl = sys.implementation

    # For CPython, use the version-specific ABI
    if impl.name.lower() == "cpython":
        major, minor = get_python_version_info()
        return f"cp{major}{minor}"

    # For other implementations, use 'none' as they typically don't have
    # version-specific ABIs in the same way
    return "none"


def get_platform_tag() -> str:
    """
    Get the platform tag for the current system.

    Returns:
        str: The platform tag
    """
    # Get the platform and normalize it
    platform_name = platform.system().lower()
    machine = platform.machine().lower()

    # Common platform mappings
    platform_mappings = {
        "linux": "linux",
        "darwin": "macosx",
        "windows": "win",
        "freebsd": "freebsd",
        "openbsd": "openbsd",
        "netbsd": "netbsd",
        "sunos": "sunos",
        "aix": "aix",
        "hp-ux": "hp-ux",
    }

    base_platform = platform_mappings.get(platform_name, platform_name)

    if base_platform == "linux":
        libc_version = platform.libc_ver()
        if libc_version is not None and libc_version[0] == "glibc":
            libc_version = libc_version[1]
            x, y = libc_version.split(".")
            base_platform = f"manylinux_{x}_{y}"
        else:
            raise ValueError(f"Unsupported libc version: {libc_version}")

    # Add architecture information
    arch_mappings = {
        "x86_64": "x86_64",
        "amd64": "x86_64",
        "i386": "i386",
        "i686": "i686",
        "aarch64": "aarch64",
        "arm64": "aarch64",
        "armv7l": "armv7l",
        "armv6l": "armv6l",
        "ppc64le": "ppc64le",
        "ppc64": "ppc64",
        "s390x": "s390x",
        "riscv64": "riscv64",
    }

    arch = arch_mappings.get(machine, machine)

    return f"{base_platform}_{arch}"


def generate_wheel_tag(
    impl: Optional[str] = None,
    major: Optional[int] = None,
    minor: Optional[int] = None,
    abi: Optional[str] = None,
    _platform: Optional[str] = None,
) -> str:
    """
    Generate a complete wheel tag.

    Args:
        impl: Implementation tag (py, cp, ip, pp, jy, etc.)
        major: Python major version
        minor: Python minor version
        abi: ABI tag
        platform: Platform tag

    Returns:
        str: Complete wheel tag in format: {python_tag}-{abi_tag}-{platform_tag}
    """
    # Use detected values if not provided
    if impl is None:
        impl = get_python_implementation_tag()

    if major is None:
        major = get_python_version_info()[0]

    if minor is None:
        minor = get_python_version_info()[1]

    if abi is None:
        abi = get_abi_tag()

    if _platform is None:
        _platform = get_platform_tag()

    # Generate the Python tag
    python_tag = f"{impl}{major}{minor}"

    # Format the complete tag
    return f"{python_tag}-{abi}-{_platform}"



import contextlib
import os
from contextlib import _GeneratorContextManager
from pathlib import Path
from typing import Callable

import pytest

from poexy_core import api
from poexy_core.pyproject.toml import PyProjectTOML


@pytest.fixture(scope="session")
def self_project() -> Path:
    project_path = Path(api.__file__).resolve().parent.parent
    return project_path


@pytest.fixture()
def sample_project(samples_path) -> Callable[[str], Path]:
    def _sample_project(name: str):
        project_path = samples_path / name
        return project_path

    return _sample_project


@pytest.fixture(scope="session")
def project() -> Callable[[Path], _GeneratorContextManager[None, None, None]]:
    @contextlib.contextmanager
    def _project(project_path: Path):
        pyproject_path = project_path / "pyproject.toml"
        if not pyproject_path.exists():
            raise FileNotFoundError(f"Pyproject.toml not found in {project_path}")

        old_cwd = os.getcwd()
        try:
            os.chdir(project_path)
            yield
        finally:
            os.chdir(old_cwd)

    return _project


@pytest.fixture()
def pyproject():
    def _pyproject():
        cwd = Path.cwd()
        return PyProjectTOML(path=cwd)

    return _pyproject



import zipfile
from pathlib import Path
from typing import Callable

import pytest
from assertpy import assert_that

from poexy_core import api
from tests.pip import Pip

# pylint: disable=redefined-outer-name


@pytest.fixture()
def pip(
    virtualenv_path, create_venv_archive, log_info_section, pyinstaller_path
) -> Pip:
    pyinstaller_path()
    log_info_section("Creating virtualenv")
    pip = Pip(virtualenv_path)
    pip.extract_virtualenv_archive(create_venv_archive)
    log_info_section("Checking poexy-core is installed")
    assert_that(pip.show("poexy-core")).is_equal_to(0)
    return pip


@pytest.fixture(scope="session")
def create_venv_archive(
    global_virtualenv_path, global_virtualenv_archive_path, self_build, log_info_section
) -> Path:
    pip = Pip(global_virtualenv_path)
    log_info_section("Creating virtualenv")
    pip.create_virtualenv()
    log_info_section("Installing poexy-core")
    pip.install(self_build)
    log_info_section("Creating virtualenv archive")
    archive_path = pip.create_virtualenv_archive(global_virtualenv_archive_path)
    assert_that(str(archive_path)).is_file()
    return archive_path


@pytest.fixture(scope="session")
def self_build(project, self_project, self_project_dist_path, log_info_section):
    with project(self_project):
        log_info_section("Building self project")
        filename = api.build_wheel(str(self_project_dist_path))
        archive_path = self_project_dist_path / filename
        assert_that(str(archive_path)).is_file()
        assert_that(zipfile.is_zipfile(archive_path)).is_true()
        return archive_path


@pytest.fixture()
def assert_pip_wheel(
    build_path, pip, dist_package_name, log_info_section
) -> Callable[[Path], Path]:
    def _assert(archive_path: Path):
        log_info_section("Running pip wheel")
        wheel_path = Path(build_path) / "wheel"
        if dist_package_name() == "poexy_core":
            no_build_isolation = False
            check_build_dependencies = True
        else:
            no_build_isolation = True
            check_build_dependencies = True
        returncode = pip.wheel(
            archive_path, wheel_path, no_build_isolation, check_build_dependencies
        )
        assert_that(returncode).is_equal_to(0)
        for file in wheel_path.iterdir():
            if file.is_file() and file.name.startswith(dist_package_name()):
                return file
        raise AssertionError(f"Wheel file {dist_package_name()} not found")

    return _assert


@pytest.fixture()
def assert_pip_install(
    install_path, pip, dist_package_name, log_info_section
) -> Callable[[Path], Path]:
    def _assert(archive_path: Path):
        log_info_section("Running pip install")
        if dist_package_name() == "poexy_core":
            no_build_isolation = False
            check_build_dependencies = True
        else:
            no_build_isolation = True
            check_build_dependencies = True
        returncode = pip.install(
            archive_path, install_path, no_build_isolation, check_build_dependencies
        )
        assert_that(returncode).is_equal_to(0)
        return install_path

    return _assert



from pathlib import Path
from typing import Any, Dict, Optional, override

from poetry.core.factory import Factory
from poetry.core.poetry import Poetry
from poetry.core.pyproject.tables import BuildSystem
from poetry.core.pyproject.toml import PyProjectTOML as PoetryPyProjectTOML
from pydantic import BaseModel, Field, field_validator

from poexy_core.pyproject.tables.license import License
from poexy_core.pyproject.tables.package import BinaryPackage, ModulePackage
from poexy_core.pyproject.tables.poexy import Poexy
from poexy_core.pyproject.tables.readme import Readme

# pylint: disable=attribute-defined-outside-init


class PyProjectTOML(BaseModel):
    path: Path = Field(description="Path to the pyproject.toml file")

    @override
    def model_post_init(self, __context: Any, /) -> None:
        self.__data: Dict[str, Any] = PoetryPyProjectTOML(self.path).data
        self.__build_system: Optional[BuildSystem] = None
        self.__poetry: Optional[Poetry] = None
        self.__poexy: Optional[Poexy] = None

    @field_validator("path")
    @classmethod
    def validate_path(cls, v: Path) -> Path:
        path = v / "pyproject.toml"
        if not path.exists():
            raise ValueError(f"Path {v} does not exist")
        if path.name != "pyproject.toml":
            raise ValueError(f"Path {v} is not a pyproject.toml file")
        return path

    @property
    def build_system(self) -> BuildSystem:
        if self.__build_system is None:
            poetry = self.poetry
            pyproject = poetry.pyproject
            package = poetry.package
            if package.name == "poexy-backend":
                dependencies = package.requires
            else:
                dependencies = poetry.build_system_dependencies
            container = pyproject.build_system
            self.__build_system = BuildSystem(
                build_backend=container.build_backend,
                requires=[dependency.to_pep_508() for dependency in dependencies],
            )

        return self.__build_system

    @property
    def poetry(self) -> Poetry:
        if self.__poetry is None:
            self.__poetry = Factory().create_poetry(self.path)
        return self.__poetry

    @property
    def poexy(self) -> Poexy:
        if self.__poexy is None:
            tool = self.__data.get("tool", None)
            if tool is None:
                tool = {}

            assert isinstance(tool, dict)

            poexy_config = tool.get("poexy", None)

            if poexy_config is None:
                poexy_config = {}

            package_config = poexy_config.get("package", None)

            if package_config is None:
                poexy_config.update(ModulePackage.from_project_config(self.__data))
            else:
                poexy_config.update(ModulePackage.from_poexy_config(poexy_config))

            binary_config = poexy_config.get("binary", None)

            if binary_config is None:
                poexy_config.update(BinaryPackage.from_project_config(self.__data))
            else:
                poexy_config.update(BinaryPackage.from_poexy_config(poexy_config))

            readme = Readme.from_project_config(self.__data)
            _license = License.from_project_config(self.__data)

            self.__poexy = Poexy(license=_license, readme=readme, **poexy_config)
        return self.__poexy



from pathlib import Path
from typing import Optional, Set

from pydantic import BaseModel, Field, model_validator

from poexy_core.packages.format import PackageFormat

# pylint: disable=no-member

SDIST_EXTENSIONS = {".py", ".c", ".cpp", ".h"}

WHEEL_EXTENSIONS = {
    ".py",
    ".pyd",
    ".so",
    ".dll",
    ".dylib",
}

GLOB_EXTENSIONS = {
    *SDIST_EXTENSIONS,
    *WHEEL_EXTENSIONS,
}

FORBIDDEN_DIRS = {
    "__pycache__",
    "build",
    "dist",
    ".eggs",
    ".mypy_cache",
    ".pytest_cache",
    ".venv",
    "venv",
}


class PackageFile(BaseModel):
    source: Path = Field(description="Source path where file will be copied from")
    destination: Path = Field(
        description="Destination path where file will be copied to"
    )

    def __eq__(self, other: "PackageFile") -> bool:
        return self.source == other.source and self.destination == other.destination

    def __ne__(self, value: "PackageFile") -> bool:
        return not self == value

    def __lt__(self, other: "PackageFile") -> bool:
        if self.destination.parent != other.destination.parent:
            return str(self.destination.parent) < str(other.destination.parent)
        return self.destination.name < other.destination.name

    def __le__(self, other: "PackageFile") -> bool:
        return self < other or self == other

    def __gt__(self, other: "PackageFile") -> bool:
        return not self <= other

    def __ge__(self, other: "PackageFile") -> bool:
        return not self < other

    def __hash__(self) -> int:
        return hash((self.source, self.destination))


PackageFiles = Set[PackageFile]


class ResolvedPackageFiles(BaseModel):
    includes: PackageFiles = Field(description="Includes files")
    excludes: PackageFiles = Field(description="Excludes files")

    @model_validator(mode="after")
    def validate_model(self) -> "ResolvedPackageFiles":
        self.includes = set(sorted(self.includes))
        self.excludes = set(self.excludes)
        return self


class ResolvePackageFiles:
    def resolve(
        self, _format: PackageFormat, base_path: Path
    ) -> Optional[PackageFiles]:
        raise NotImplementedError("Subclass must implement this method")


class FilePattern(BaseModel):
    glob_pattern: Optional[Path] = Field(
        description="Glob pattern to match the files in the built archive"
    )
    path: Path = Field(description="Path where to match the files")

    @model_validator(mode="after")
    def validate_model(self) -> "FilePattern":
        if self.glob_pattern is not None and self.glob_pattern.is_absolute():
            raise ValueError("Glob pattern must be relative")
        if self.glob_pattern is not None and "*" not in str(self.glob_pattern):
            raise ValueError("Invalid glob pattern")
        if self.glob_pattern is None and not self.path.is_file():
            raise ValueError("Glob pattern is required for directories")
        return self

    def resolve(self, destination_path: Path) -> PackageFiles:
        glob_pattern = self.glob_pattern
        source_path = self.path.resolve()

        base_path = Path.cwd()
        source_path = source_path.relative_to(base_path)

        def normalize_path(path: Path) -> PackageFile:
            try:
                relative_path = path.relative_to(base_path)
                return PackageFile(
                    source=relative_path,
                    destination=destination_path / relative_path.relative_to(self.path),
                )
            except ValueError:
                relative_path = Path(base_path / path).relative_to(base_path)
                return PackageFile(
                    source=relative_path,
                    destination=destination_path / relative_path,
                )

        if glob_pattern is None and source_path.is_file():
            return {
                PackageFile(
                    source=source_path,
                    destination=source_path,
                )
            }

        files = []

        str_glob_pattern = str(glob_pattern)
        str_glob_pattern = f"{source_path}/{str_glob_pattern}"

        for source_path in base_path.rglob(str_glob_pattern):
            if not source_path.is_file():
                continue
            if any(part in FORBIDDEN_DIRS for part in source_path.parts):
                continue
            relative_path = normalize_path(source_path)
            files.append(relative_path)

        if len(files) == 0:
            raise ValueError("No files found")
        return set(files)



from pathlib import Path
from typing import override

from poetry.core.masonry.metadata import Metadata
from poetry.core.poetry import Poetry

from poexy_core.builders.hooks.hook import HookBuilder
from poexy_core.builders.types import FilePathCallback
from poexy_core.packages.format import PackageFormat
from poexy_core.pyproject.tables.poexy import Poexy


class ReadmeHookBuilder(HookBuilder):
    def __init__(self, poetry: Poetry, poexy: Poexy, destination_path: Path):
        super().__init__("readme")
        self.__poetry = poetry
        self.__poexy = poexy
        self.__destination_path = destination_path
        self.__files = []

    @override
    def build(self):
        with self._hook_build():
            if self.__poexy.readme is None:
                return
            self.__poexy.readme.transform_poetry_package(self.__poetry.package)
            self.__files.append(
                (
                    self.__poexy.readme.file.pattern,
                    self.__destination_path / self.__poexy.readme.file.pattern,
                )
            )

    @override
    def add_files(self, callback: FilePathCallback):
        if self.__poexy.readme is None:
            return
        for file in self.__files:
            callback(file[0], file[1])

    @override
    def add_metadata(self, metadata: Metadata, _format: PackageFormat):
        pass



from pathlib import Path

from assertpy import assert_that

from poexy_core import api


def test_get_requires_for_build_wheel():
    requires = api.get_requires_for_build_wheel()
    assert_that(requires).is_not_empty()


def test_prepare_metadata_for_build_wheel(tmp_path, dist_package_name, package_version):
    metadata_name = api.prepare_metadata_for_build_wheel(str(tmp_path))

    dist_package_name = dist_package_name()
    package_version = package_version()
    assert_that(metadata_name).is_equal_to(
        f"{dist_package_name}-{package_version}.dist-info"
    )


def test_wheel(
    self_project,
    dist_package_name,
    default_python_tag,
    wheel_data_purelib_folder,
    assert_wheel_build,
):
    assert_zip_file = assert_wheel_build(self_project)
    assert_zip_file(
        [
            wheel_data_purelib_folder(default_python_tag)
            / dist_package_name()
            / "api.py",
        ]
    )


def test_sdist(
    self_project,
    dist_package_name,
    site_packages_path,
    assert_sdist_build,
):
    assert_tar_file = assert_sdist_build(self_project)
    assert_tar_file(
        [
            Path(dist_package_name()) / "api.py",
            Path("tests") / "conftest.py",
        ]
    )

    site_packages = site_packages_path / dist_package_name()
    assert_that(str(site_packages)).is_directory()



import tarfile
import zipfile
from pathlib import Path
from typing import Callable, List, Set

import pytest
from assertpy import assert_that

from poexy_core import api
from poexy_core.builders.wheel import WheelMetadata
from poexy_core.packages.format import WheelFormat

# pylint: disable=redefined-outer-name,dangerous-default-value

__all__ = [
    "assert_tar_file",
    "assert_zip_file",
    "assert_wheel_build",
    "assert_sdist_build",
]


@pytest.fixture()
def assert_tar_file() -> Callable[[Path, List[Path], bool], None]:
    def _assert(archive_path: Path, expected_files: List[Path], strict: bool = False):
        with tarfile.open(archive_path, "r:gz") as tar:
            members = tar.getmembers()
            assert_that(members).is_not_empty()

            archive_file_paths = [Path(member.name) for member in members]

            for member in archive_file_paths:
                if member.name == "PKG-INFO":
                    root_folder = member.parent
                    break

            assert_that(root_folder).is_not_none()

            archive_file_paths = [
                Path(member.name).relative_to(root_folder) for member in members
            ]

            if strict:
                # Assert that all expected files are present and no extra files exist
                archive_file_paths = [
                    path
                    for path in archive_file_paths
                    if path.name not in ["PKG-INFO", "pyproject.toml"]
                ]
                assert_that(len(archive_file_paths)).is_equal_to(len(expected_files))
                assert_that(sorted(archive_file_paths)).is_equal_to(
                    sorted(expected_files)
                )
            else:
                # Assert that all expected files are present (partial check)
                for expected_file in expected_files:
                    assert_that(archive_file_paths).contains(expected_file)

    return _assert


@pytest.fixture()
def assert_zip_file() -> Callable[[Path, List[Path], bool], None]:
    def _assert(
        archive_path: Path,
        expected_files: List[Path],
        strict: bool = False,
        strip: bool = True,
    ):
        with zipfile.ZipFile(archive_path, "r") as zip_file:
            members = zip_file.namelist()
            assert_that(members).is_not_empty()

            archive_file_paths = [Path(member) for member in members]

            if strict:
                # Assert that all expected files are present and no extra files exist
                if strip:
                    archive_file_paths = [
                        path
                        for path in archive_file_paths
                        if path.name not in ["WHEEL", "METADATA", "RECORD"]
                    ]
                assert_that(len(archive_file_paths)).is_equal_to(len(expected_files))
                assert_that(sorted(archive_file_paths)).is_equal_to(
                    sorted(expected_files)
                )
            else:
                # Assert that all expected files are present (partial check)
                for expected_file in expected_files:
                    assert_that(archive_file_paths).contains(expected_file)

    return _assert


@pytest.fixture()
def assert_wheel_build(
    project,
    dist_path,
    install_path,
    dist_package_name,
    package_name,
    site_packages_path,
    current_python_tag,
    default_python_tag,
    wheel_metadata,
    wheel_data_scripts_folder,
    assert_metadata_manifest,
    assert_wheel_manifest,
    assert_record_manifest,
    assert_pip_install,
    assert_zip_file,
) -> Callable[[Path], Callable[[List[Path], bool], None]]:
    def _assert(project_path: Path, _format: Set[WheelFormat] = {WheelFormat.Source}):
        with project(project_path):
            filename = api.build_wheel(str(dist_path))

            assert_that(filename).ends_with(".whl")
            assert_that(filename).starts_with(dist_package_name())
            archive_path = dist_path / filename
            assert_that(str(archive_path)).is_file()
            assert_that(zipfile.is_zipfile(archive_path)).is_true()

            if len(_format) == 1 and WheelFormat.Binary in _format:
                python_tag = current_python_tag
            else:
                python_tag = default_python_tag

            metadata: WheelMetadata = wheel_metadata(python_tag)
            assert_metadata_manifest(python_tag)
            assert_wheel_manifest(python_tag)
            assert_record_manifest(python_tag)
            assert_pip_install(archive_path)

            site_packages = site_packages_path / dist_package_name()

            if len(_format) == 1 and WheelFormat.Binary in _format:
                assert_that(str(site_packages)).does_not_exist()
                binary_path = install_path / "bin" / package_name()
                assert_that(binary_path.exists()).is_true()
                dist_info_folder = metadata.dist_info_folder.name
                assert_zip_file(
                    archive_path,
                    [
                        Path(dist_info_folder) / "WHEEL",
                        Path(dist_info_folder) / "METADATA",
                        Path(dist_info_folder) / "RECORD",
                        wheel_data_scripts_folder(python_tag) / package_name(),
                    ],
                    strict=True,
                    strip=False,
                )
            else:
                assert_that(str(site_packages)).exists()
                binary_path = install_path / "bin" / package_name()
                if WheelFormat.Binary in _format:
                    assert_that(binary_path.exists()).is_true()
                else:
                    assert_that(binary_path.exists()).is_false()
                dist_info_folder = metadata.dist_info_folder.name
                expected_zip_files = [
                    Path(dist_info_folder) / "WHEEL",
                    Path(dist_info_folder) / "METADATA",
                    Path(dist_info_folder) / "RECORD",
                ]
                if WheelFormat.Binary in _format:
                    expected_zip_files.append(
                        wheel_data_scripts_folder(python_tag) / package_name()
                    )
                assert_zip_file(archive_path, expected_zip_files)

            def _assert_zip_file(expected_files: List[Path], strict: bool = False):
                assert_zip_file(archive_path, expected_files, strict)

            return _assert_zip_file

    return _assert


@pytest.fixture()
def assert_sdist_build(
    project,
    dist_path,
    site_packages_path,
    install_path,
    dist_package_name,
    package_name,
    assert_pkginfo_manifest,
    assert_pip_wheel,
    assert_pip_install,
    assert_tar_file,
) -> Callable[[Path], Callable[[List[Path], bool], None]]:
    def _assert(project_path: Path, _format: Set[WheelFormat] = {WheelFormat.Source}):
        with project(project_path):
            filename = api.build_sdist(str(dist_path))

            assert_that(filename).ends_with(".tar.gz")
            assert_that(filename).starts_with(dist_package_name())
            archive_path = dist_path / filename
            assert_that(str(archive_path)).is_file()
            assert_that(tarfile.is_tarfile(archive_path)).is_true()

            assert_pkginfo_manifest()
            wheel_path = assert_pip_wheel(archive_path)
            assert_pip_install(wheel_path)

            site_packages = site_packages_path / dist_package_name()

            if len(_format) == 1 and WheelFormat.Binary in _format:
                assert_that(str(site_packages)).does_not_exist()
                binary_path = install_path / "bin" / package_name()
                assert_that(binary_path.exists()).is_true()
            else:
                assert_that(str(site_packages)).exists()
                binary_path = install_path / "bin" / package_name()
                if WheelFormat.Binary in _format:
                    assert_that(binary_path.exists()).is_true()
                else:
                    assert_that(binary_path.exists()).is_false()

            assert_tar_file(archive_path, [Path("pyproject.toml"), Path("PKG-INFO")])

            def _assert_tar_file(expected_files: List[Path], strict: bool = False):
                assert_tar_file(archive_path, expected_files, strict)

            return _assert_tar_file

    return _assert



import csv
import hashlib
import logging
from pathlib import Path
from typing import Callable, List, Optional

from poexy_core.manifest.parser import parse, to_string
from poexy_core.manifest.types import ManifestKey, ManifestStorage, ManifestValue
from poexy_core.metadata.fields import MetadataField

logger = logging.getLogger(__name__)


class Manifest:
    def __init__(self, path: Path):
        self.path = path
        self.__storage: ManifestStorage = []

        if not self.path.parent.exists():
            self.path.parent.mkdir(parents=True, exist_ok=True)

    def __len__(self):
        return len(self.__storage)

    def __iter__(self):
        return iter(self.__storage)

    def __getitem__(self, key: ManifestKey) -> ManifestValue:
        return self.get(key)

    def __setitem__(self, key: ManifestKey, value: ManifestValue):
        self.set(key, value)

    def __delitem__(self, key: ManifestKey):
        self.delete(key)

    def read(self):
        logger.info(f"Reading {self.path.name} manifest")
        with open(self.path, "r", encoding="utf-8") as file:
            content = file.read()
            self.__storage = parse(content)

    def write(self):
        logger.info(f"Writing {self.path.name} manifest")
        with open(self.path, "w", encoding="utf-8") as file:
            file.write(to_string(self.__storage))

    def get(self, key: ManifestKey) -> ManifestValue:
        logger.info(f"Getting [{self.path.name}] {key}")
        if isinstance(key, MetadataField):
            key = key.name
        for k, v in self.__storage:
            if k == key:
                return v
        raise KeyError(key)

    def set(self, key: ManifestKey, value: ManifestValue):
        if isinstance(value, list) or isinstance(value, tuple):
            log_value = "\n  - ".join(value)
            logger.info(f"Setting [{self.path.name}] {key}: \n  - {log_value}")
        else:
            logger.info(f"Setting [{self.path.name}] {key}: {value}")
        if isinstance(key, MetadataField):
            key = key.name
        for k, _ in self.__storage:
            if k == key:
                raise KeyError(key)
        self.__storage.append((key, value))

    def delete(self, key: ManifestKey):
        logger.info(f"Deleting [{self.path.name}] {key}")
        if isinstance(key, MetadataField):
            key = key.name
        for i, (k, _) in enumerate(self.__storage):
            if k == key:
                del self.__storage[i]
                return
        raise KeyError(key)


ManifestOperation = Callable[[Manifest], None]


class MetadataManifest(Manifest):
    def __init__(self, path: Path):
        super().__init__(path / "METADATA")


class WheelManifest(Manifest):
    def __init__(self, path: Path):
        super().__init__(path / "WHEEL")


class PackageInfoManifest(Manifest):
    def __init__(self, path: Path):
        super().__init__(path / "PKG-INFO")


class Record:
    def __init__(self, path: Path, sha: Optional[str], size: Optional[int]):
        self.path = path
        self.sha = sha
        self.size = size

    def to_relative_path(self, base_path: Path):
        self.path = self.path.relative_to(base_path)

    @staticmethod
    def from_path(path: Path) -> "Record":
        if not path.is_file():
            raise FileNotFoundError(f"{path} is not a file")
        if not path.exists():
            raise FileNotFoundError(f"{path} does not exist")
        return Record(
            path,
            f"sha256={hashlib.sha256(path.read_bytes()).hexdigest()}",
            path.stat().st_size,
        )

    @staticmethod
    def from_paths(source: Path, destination: Path) -> "Record":
        if destination.exists():
            raise FileExistsError(f"{destination} should not exist")
        record = Record.from_path(source)
        record.path = destination
        return record


ManifestRecordStorage = List[Record]
ManifestRecordStorageOperation = Callable[[ManifestRecordStorage], None]


class RecordManifest:
    def __init__(self, path: Path):
        self.__base_path = path.parent  # ./dist-info
        self.path = path / "RECORD"
        self.__storage: ManifestRecordStorage = []

        if not self.path.parent.exists():
            self.path.parent.mkdir(parents=True, exist_ok=True)

    def __len__(self):
        return len(self.__storage)

    def __iter__(self):
        return iter(self.__storage)

    def read(self):
        logger.info(f"Reading {self.path.name} manifest")
        with open(self.path, "r", encoding="utf-8") as file:
            reader = csv.reader(file, lineterminator="\n")
            for row in reader:
                path = Path(row[0])
                sha = row[1]
                size = int(row[2]) if len(row[2]) > 0 else None
                self.__storage.append(Record(path, sha, size))

    def __write(self):
        for record in self.__storage:
            record.to_relative_path(self.__base_path)
        with open(self.path, "w", encoding="utf-8") as file:
            writer = csv.writer(file, lineterminator="\n")
            rows = []
            for record in self.__storage:
                size = record.size if record.size is not None else ""
                row = [str(record.path), record.sha, size]
                rows.append(row)
            writer.writerows(rows)

    def write(self):
        logger.info(f"Writing {self.path.name} manifest")
        self.__write()

    def add(self, path: Path):
        logger.info(f"Adding [{self.path.name}]: {path.relative_to(self.__base_path)}")
        self.__storage.append(Record.from_path(path))

    def add_self(self):
        logger.info(
            f"Adding [{self.path.name}] self: "
            f"{self.path.relative_to(self.__base_path)}"
        )
        self.__storage.append(Record(self.path, None, None))

    def set(self, source: Path, destination: Path):
        logger.info(
            f"Setting [{self.path.name}]: {destination.relative_to(self.__base_path)}"
        )
        for record in self.__storage:
            if record.path == destination:
                raise KeyError(destination)
        self.__storage.append(Record.from_paths(source, destination))



from typing import Optional

from poexy_core.manifest.types import ManifestStorage

MULTILINE_PREFIX = " " * 7 + "|"


def __get_storage_item(storage: ManifestStorage, key: str) -> Optional[int]:
    for index, item in enumerate(storage):
        if item[0] == key:
            return index
    return None


def _parse_multiline_value(
    storage: ManifestStorage, item_key: Optional[str], line: str
):
    if item_key is None:
        raise ValueError("Not key found for multiline value")
    storage_item_index = __get_storage_item(storage, item_key)
    if storage_item_index is None:
        raise ValueError(f"Key {item_key} not found in storage")
    storage_item_key, storage_item_value = storage[storage_item_index]
    if isinstance(storage_item_value, str):
        if MULTILINE_PREFIX in line:
            line = line.replace(MULTILINE_PREFIX, "").strip()
        storage[storage_item_index] = (
            storage_item_key,
            storage_item_value + "\n" + line,
        )
    else:
        raise ValueError(f"Key {item_key} multiline value is not a string")


def __parse_single_line_value(storage: ManifestStorage, line: str) -> str:
    key, value = line.split(":", 1)
    storage_item_index = __get_storage_item(storage, key)
    if storage_item_index is None:
        storage.append((key, value.strip()))
    else:
        storage_item_key, storage_item_value = storage[storage_item_index]
        if isinstance(storage_item_value, str):
            storage_item_value = [storage_item_value]
        if isinstance(storage_item_value, list):
            storage[storage_item_index] = (
                storage_item_key,
                [*storage_item_value, value],
            )
        elif isinstance(storage_item_value, tuple):
            storage[storage_item_index] = (
                storage_item_key,
                (*storage_item_value, value),
            )
        else:
            raise ValueError(f"Unknown value type: {type(storage_item_value)}")
    return key


def parse(text: str) -> ManifestStorage:
    storage: ManifestStorage = []
    last_key: Optional[str] = None
    for line in text.splitlines():
        if line.strip() == "":
            continue
        if line.startswith(MULTILINE_PREFIX):
            _parse_multiline_value(storage, last_key, line)
        else:
            last_key = __parse_single_line_value(storage, line)
    return storage


def to_string(storage: ManifestStorage) -> str:
    text = ""
    for key, value in storage:
        if isinstance(value, str):
            text += f"{key}: {value}\n"
        elif isinstance(value, list):
            for item in value:
                text += f"{key}: {item}\n"
        elif isinstance(value, tuple):
            for item in value:
                text += f"{key}: {item}\n"
        else:
            raise ValueError(f"Unknown value type: {type(value)}")
    return text



import logging
from pathlib import Path
from typing import Any, Dict, List, Optional

from poetry.core.masonry.metadata import Metadata
from pydantic import BaseModel, Field, model_validator

from poexy_core.packages.format import PackageFormat
from poexy_core.pyproject.exceptions import PyProjectError
from poexy_core.pyproject.types import GlobPattern

logger = logging.getLogger(__name__)


class License(BaseModel):
    text: Optional[str] = Field(description="License text", default=None)
    expression: Optional[str] = Field(description="License expression", default=None)
    files: Optional[List[GlobPattern]] = Field(
        description="License files", default=None
    )

    @model_validator(mode="after")
    def validate_model(self) -> "License":
        if self.text is not None and self.expression is not None:
            raise ValueError("text and expression cannot be used together")
        if self.text is not None and self.files is not None:
            raise ValueError("text and files cannot be used together")
        if self.expression is not None and self.files is not None:
            raise ValueError("expression and files cannot be used together")
        return self

    def resolve_files(self) -> Optional[List[Path]]:
        if self.files is None:
            return None
        current_working_directory = Path.cwd()
        if not (current_working_directory / "pyproject.toml").exists():
            raise PyProjectError(
                "License files must be resolved from the project root directory"
            )
        files = []
        for file in self.files:
            resolved_files = file.resolve()
            if resolved_files is not None:
                files.extend(resolved_files)
        return files

    @staticmethod
    def from_project_config(config: Dict[str, Any]) -> Optional["License"]:
        project = config.get("project", None)
        assert isinstance(project, dict)
        license_files = project.get("license-files", None)
        _license = project.get("license", None)
        if isinstance(_license, Dict) and len(_license) == 1:
            license_text = _license.get("text", None)
            if license_text is not None:
                logger.warning(
                    "Key 'text' in 'license' table is deprecated, use 'license' table "
                    "directly instead"
                )
                return License(text=license_text)
            license_file = _license.get("file", None)
            if license_file is not None:
                logger.warning(
                    "Key 'file' in 'license' table is deprecated, use 'license-files'"
                    "table instead"
                )
                return License(files=[GlobPattern(pattern=license_file)])
            raise PyProjectError("[project] section must contain a valid license table")
        if isinstance(license_files, List):
            return License(files=[GlobPattern(pattern=file) for file in license_files])
        elif isinstance(_license, str):
            return License(expression=_license)
        else:
            return None

    def transform_poetry_metadata(self, metadata: Metadata, _format: PackageFormat):
        metadata.license = None
        license_files = self.resolve_files()
        if license_files is not None:
            if _format == PackageFormat.Wheel:
                file_prefix = "licences/"
            else:
                file_prefix = ""
            setattr(
                metadata,
                "license_files",
                [f"{file_prefix}{file}" for file in license_files],
            )
        if self.text is not None:
            setattr(metadata, "license", self.text)
        if self.expression is not None:
            setattr(metadata, "license_expression", self.expression)



import json
import os
from pathlib import Path
from typing import Optional

from poexy_core.utils import subprocess_rt


class Pip:
    def __init__(self, virtualenv_path: Path):
        self.__virtualenv_path = virtualenv_path
        self.__virtualenv_path.mkdir(parents=True, exist_ok=True)
        self.__cache_path = virtualenv_path / ".cache"
        self.__cache_path.mkdir(parents=True, exist_ok=True)
        self.__common_args = [
            # "--verbose",
            "--require-virtualenv",
            "--isolated",
            "--cache-dir",
            str(self.__cache_path),
        ]

    def __create_virtualenv(self) -> int:
        cmd = [
            "poetry",
            "run",
            "virtualenv",
            "--always-copy",
            "--seeder",
            "app-data",
            "--no-download",
            "--no-setuptools",
            str(self.__virtualenv_path),
        ]
        exit_code = subprocess_rt.run(cmd, printer=print)
        return exit_code

    def __get_pip_path(self) -> Path:
        return self.__virtualenv_path / "bin" / "pip"

    def __execute_pip_command(self, cmd: list[str]) -> int:
        cmd = [str(self.__get_pip_path()), *cmd]
        env = os.environ.copy()
        exit_code = subprocess_rt.run(cmd, printer=print, env=env)
        return exit_code

    def __upgrade_scripts_shebang(self) -> None:
        with open(self.__virtualenv_path / "venv.json", "r", encoding="utf-8") as f:
            venv_config = json.load(f)

        bin_path = self.__virtualenv_path / "bin"
        if not bin_path.exists():
            return

        binary_files = [
            f for f in bin_path.iterdir() if f.is_file() and f.stat().st_mode & 0o111
        ]

        for file in binary_files:
            try:
                content = file.read_text(encoding="utf-8")
            except UnicodeDecodeError:
                continue
            if venv_config["base_path"] in content:
                file.write_text(
                    content.replace(
                        venv_config["base_path"], str(self.__virtualenv_path)
                    ),
                    encoding="utf-8",
                )

    def create_virtualenv(self) -> None:
        exit_code = self.__create_virtualenv()
        if exit_code != 0:
            raise RuntimeError(f"Failed to create virtualenv: {exit_code}")

    def create_virtualenv_archive(self, archive_path: Path) -> Path:
        archive_name = "venv.tar.zst"

        venv_config = {
            "base_path": str(self.__virtualenv_path),
            "created_at": str(Path().stat().st_mtime),
        }
        with open(self.__virtualenv_path / "venv.json", "w", encoding="utf-8") as f:
            json.dump(venv_config, f)

        cmd = [
            "tar",
            "--use-compress-program",
            "zstd",
            "--create",
            "--file",
            str(archive_path / archive_name),
            "--directory",
            str(self.__virtualenv_path),
            ".",
        ]
        exit_code = subprocess_rt.run(cmd, printer=print)
        if exit_code != 0:
            raise RuntimeError(f"Failed to create virtualenv archive: {exit_code}")
        return archive_path / archive_name

    def extract_virtualenv_archive(self, archive_path: Path) -> None:
        cmd = [
            "rm",
            "-rf",
            str(self.__virtualenv_path),
        ]
        exit_code = subprocess_rt.run(cmd, printer=print)
        if exit_code != 0:
            raise RuntimeError(f"Failed to remove virtualenv: {exit_code}")
        self.__virtualenv_path.mkdir(parents=True, exist_ok=True)
        cmd = [
            "tar",
            "--use-compress-program",
            "zstd",
            "--extract",
            "--file",
            str(archive_path),
            "--directory",
            str(self.__virtualenv_path),
        ]
        exit_code = subprocess_rt.run(cmd, printer=print)

        if exit_code != 0:
            raise RuntimeError(f"Failed to extract virtualenv archive: {exit_code}")

        self.__upgrade_scripts_shebang()

    def wheel(
        self,
        archive_path: Path,
        wheel_directory: Optional[Path] = None,
        no_build_isolation: bool = True,
        check_build_dependencies: bool = True,
    ) -> int:
        cmd = ["wheel"]
        if wheel_directory is not None:
            cmd.extend(["--wheel-dir", str(wheel_directory)])
        if no_build_isolation:
            cmd.append("--no-build-isolation")
        if check_build_dependencies:
            cmd.append("--check-build-dependencies")
        cmd.extend(["--use-pep517", "--no-clean"])
        cmd.extend(self.__common_args)
        cmd.append(str(archive_path))
        return self.__execute_pip_command(cmd)

    def install(
        self,
        archive_path: Path,
        install_directory: Optional[Path] = None,
        no_build_isolation: bool = True,
        check_build_dependencies: bool = True,
    ) -> int:
        cmd = [
            "install",
        ]
        if install_directory is not None:
            cmd.append("--prefix")
            cmd.append(str(install_directory))
        if no_build_isolation:
            cmd.append("--no-build-isolation")
        if check_build_dependencies:
            cmd.append("--check-build-dependencies")
            cmd.extend(self.__common_args)
        cmd.extend(["--use-pep517", "--no-clean", "--force-reinstall"])
        cmd.extend(self.__common_args)
        cmd.append(str(archive_path))
        return self.__execute_pip_command(cmd)

    def uninstall(self, package_name: str) -> int:
        cmd = ["uninstall", "--yes"]
        cmd.extend(self.__common_args)
        cmd.append(package_name)
        return self.__execute_pip_command(cmd)

    def show(self, package_name: str) -> int:
        cmd = ["show"]
        cmd.extend(self.__common_args)
        cmd.append(package_name)
        return self.__execute_pip_command(cmd)



import glob
import re
from pathlib import Path
from typing import List, Optional, Tuple

from pydantic import BaseModel, Field, model_validator

from poexy_core.pyproject.exceptions import PyProjectError


class GlobPattern(BaseModel):
    pattern: Path = Field(description="Glob pattern")

    @model_validator(mode="after")
    def validate_model(self) -> "GlobPattern":
        pattern = str(self.pattern)
        if self.pattern.is_absolute():
            raise ValueError(f"Pattern {pattern} should be relative")
        if ".." in pattern:
            raise ValueError(f"Pattern {pattern} cannot contain '..'")
        if re.match(r"^[_\w\-\.\/\*\?\[\]]+$", pattern) is None:
            raise ValueError(f"Pattern '{pattern}' contains invalid characters.")
        self.pattern = Path(self.pattern.as_posix())
        return self

    def resolve(self) -> Optional[List[Path]]:
        current_working_directory = Path.cwd()
        if not (current_working_directory / "pyproject.toml").exists():
            raise PyProjectError(
                "Files must be resolved from the project root directory"
            )
        files = []
        for file in glob.glob(str(self.pattern)):
            path = Path(file).resolve()
            path = path.relative_to(current_working_directory)
            if path.is_file():
                files.append(path)
        if len(files) == 0:
            return None
        return files

    def split(self) -> Tuple[Path, Optional[Path]]:
        """
        Splits the glob pattern into a path and a glob pattern.
        """
        str_path = str(self.pattern)
        path = []
        for char in str_path:
            if char == "*":
                break
            path.append(char)
        glob_pattern = str_path[len(path) :]  # noqa: E203
        path = Path("".join(path))
        if path.is_absolute():
            path = path.relative_to(path.parent)
        if len(glob_pattern) == 0 and path.is_file():
            glob_pattern = None
        elif len(glob_pattern) == 0 and path.is_dir():
            glob_pattern = Path("**/*")
        else:
            glob_pattern = Path(glob_pattern)
        return path, glob_pattern



from enum import Enum
from pathlib import Path
from typing import Any, Dict, Optional

from poetry.core.packages.package import Package
from pydantic import BaseModel, Field, model_validator

from poexy_core.pyproject.types import GlobPattern

# pylint: disable=no-member


class ReadmeContentType(str, Enum):
    Markdown = "text/markdown"
    RestructuredText = "text/x-rst"


class Readme(BaseModel):
    file: GlobPattern = Field(
        description="The file to use as the README for the project.",
    )
    content_type: Optional[ReadmeContentType] = Field(
        description="The content type of the README file.",
    )

    @model_validator(mode="before")
    @classmethod
    def validate_model(cls, data: Any) -> Any:
        if not isinstance(data, dict):
            raise ValueError(f"Invalid readme type: {type(data)}")
        file = data.get("file", None)
        if file is None:
            raise ValueError("File is required")
        if isinstance(file, str):
            file = GlobPattern(pattern=Path(file))
        content_type = data.get("content-type", None)
        if content_type is not None:
            content_type = ReadmeContentType(content_type)
        if file.pattern.suffix == ".md":
            content_type = ReadmeContentType.Markdown
        elif file.pattern.suffix == ".rst":
            content_type = ReadmeContentType.RestructuredText
        else:
            raise ValueError(
                "Cannot determine content type from file extension: "
                f"{file.pattern.suffix}"
            )
        data["file"] = file
        data["content_type"] = content_type
        return data

    @staticmethod
    def from_project_config(config: Dict[str, Any]) -> Optional["Readme"]:
        project = config.get("project", None)
        assert isinstance(project, dict)
        readme = project.get("readme", None)
        if readme is None:
            return None
        if isinstance(readme, str):
            return Readme(file=GlobPattern(pattern=Path(readme)), content_type=None)
        elif isinstance(readme, dict):
            return Readme(**readme)
        else:
            raise ValueError(f"Invalid readme type: {type(readme)}")

    def transform_poetry_package(self, package: Package):
        package.readmes = ()
        readme_content = self.file.pattern.read_text(encoding="utf-8")
        package.readme_content = readme_content
        if self.content_type is None:
            raise ValueError("Content type is required")
        package.readme_content_type = self.content_type.value



import logging
import shutil
import sys
from contextlib import contextmanager
from pathlib import Path
from typing import Any, Generator, Tuple, override

from poetry.core.poetry import Poetry

from poexy_core.builders.builder import PythonTag
from poexy_core.builders.wheel import Manifests, WheelBuilder
from poexy_core.metadata.fields import MetadataField
from poexy_core.packages.format import PackageFormat
from poexy_core.pyinstaller.builder import BuildType, PyinstallerBuilder
from poexy_core.pyproject.tables.poexy import Poexy

logger = logging.getLogger(__name__)


class BinaryBuilder(WheelBuilder):
    def __init__(
        self,
        poetry: Poetry,
        poexy: Poexy,
        _format: PackageFormat,
        wheel_directory: Path,
        metadata_directory: Path | None = None,
        config_settings: dict[str, Any] | None = None,
    ):
        if _format != PackageFormat.Binary:
            raise ValueError(f"Invalid format: {_format}")
        super().__init__(
            poetry,
            poexy,
            PackageFormat.Wheel,
            wheel_directory,
            metadata_directory,
            config_settings,
        )
        python_tag = PythonTag.from_current_environment()
        self._init_metadata(python_tag)
        self._manifests = Manifests(self._metadata.dist_info_folder)

    def __add_executable_file(self, executable: Tuple[Path, Path]):
        def add_file(source: Path, destination: Path):
            logger.info(f"Recording file: {destination}")
            self._manifests.record.set(source, destination)

        add_file(executable[0], executable[1])
        self._add_files_to_archive(executable[0], executable[1])

    @override
    def _add_wheel(self):
        self._manifests.wheel.set("Wheel-Version", "1.0")
        self._manifests.wheel.set("Generator", "Poexy")
        self._manifests.wheel.set("Root-Is-Purelib", "false")
        self._manifests.wheel.set("Tag", f"{self._metadata.python_tag}")

    def __add_metadata(self):
        builder = self._add_metadata(self._manifests.metadata)
        classifiers = list(builder.get(MetadataField.Classifier))
        classifiers = [
            c
            for c in classifiers
            if not c.startswith("Programming Language :: Python ::")
        ]
        builder.delete(MetadataField.Classifier)
        classifiers.append(
            " :: ".join(
                [
                    "Programming Language",
                    "Python",
                    f"{sys.version_info.major}.{sys.version_info.minor}",
                ]
            )
        )
        classifiers.append("Operating System :: OS Independent")
        builder.set(MetadataField.Classifier, classifiers)
        fields_to_delete = [
            MetadataField.RequiresDist,
            MetadataField.RequiresExternal,
            MetadataField.RequiresPython,
            MetadataField.ProvidesExtra,
            MetadataField.ProvidesDist,
            MetadataField.SupportedPlatforms,
        ]
        for field in fields_to_delete:
            try:
                builder.delete(field)
            except KeyError:
                pass
        return builder

    def prepare(self) -> Path:
        logger.info("Preparing wheel metadata...")
        logger.info("Adding metadata...")
        self.__add_metadata()
        logger.info("Adding wheel...")
        self._add_wheel()
        logger.info("Writing manifests...")
        self._manifests.metadata.write()
        self._manifests.wheel.write()
        logger.info("Wheel metadata prepared successfully.")
        return self._metadata.dist_info_folder

    def build_executable(self) -> Tuple[Path, Path]:
        pyinstaller_builder = PyinstallerBuilder(self.poetry, self.poexy)
        logger.info(f"Building executable: {pyinstaller_builder.executable_name}")
        pyinstaller_builder.build(
            build_type=BuildType.OneFile,
            dist_path=self.destination_directory,
            strip=True,
            clean=True,
        )
        executable_name = pyinstaller_builder.executable_name
        source = self.destination_directory / executable_name
        destination = self._metadata.data_scripts_folder / executable_name
        return source, destination

    @override
    @contextmanager
    def build(self) -> Generator[Path, None, None]:
        logger.info("Building binary...")

        executable = self.build_executable()

        for hook in self._hooks:
            hook.build()

        with self._create_archive():
            logger.info("Adding metadata...")
            self._add_metadata(self._manifests.metadata)
            logger.info("Adding wheel...")
            self._add_wheel()
            logger.info("Adding executable...")
            self.__add_executable_file(executable)
            logger.info("Writing manifests...")
            self._manifests.write()
            logger.info("Adding dist-info files...")
            self._add_dist_info_files_to_archive()

        logger.info("Built successfully.")

        shutil.copy(self._metadata.archive_path, self.destination_directory)

        yield self._metadata.archive_path



from pathlib import Path

import pytest
from assertpy import assert_that

from poexy_core.packages.format import WheelFormat

# pylint: disable=redefined-outer-name


@pytest.fixture()
def project_path(sample_project):
    return sample_project("wheel_binary")


def test_wheel(
    project,
    project_path,
    assert_wheel_build,
    wheel_data_scripts_folder,
    current_python_tag,
    dist_package_name,
    package_name,
    site_packages_path,
    install_path,
):
    with project(project_path):
        assert_zip_file = assert_wheel_build(project_path, _format={WheelFormat.Binary})
        assert_zip_file(
            [
                wheel_data_scripts_folder(current_python_tag) / package_name(),
            ],
            strict=True,
        )
        purelib_path = site_packages_path / dist_package_name() / "__init__.py"
        assert_that(purelib_path.exists()).is_false()
        binary_path = install_path / "bin" / package_name()
        assert_that(binary_path.exists()).is_true()


def test_sdist(
    project,
    project_path,
    assert_sdist_build,
    dist_package_name,
    package_name,
    site_packages_path,
    install_path,
):
    with project(project_path):
        assert_tar_file = assert_sdist_build(project_path, _format={WheelFormat.Binary})
        assert_tar_file(
            [
                Path("src") / "__init__.py",
            ],
            strict=True,
        )
        purelib_path = site_packages_path / dist_package_name() / "__init__.py"
        assert_that(purelib_path.exists()).is_false()
        binary_path = install_path / "bin" / package_name()
        assert_that(binary_path.exists()).is_true()



import logging
import shutil
import tarfile
from contextlib import contextmanager
from pathlib import Path
from typing import Any, Generator, Optional, override

from packaging.utils import NormalizedName
from poetry.core.masonry.utils.helpers import distribution_name
from poetry.core.poetry import Poetry

from poexy_core.builders.builder import Builder
from poexy_core.builders.hooks.include_files import IncludeFilesHookBuilder
from poexy_core.builders.hooks.license import LicenseHookBuilder
from poexy_core.builders.hooks.package_files import PackageFilesHookBuilder
from poexy_core.builders.hooks.readme import ReadmeHookBuilder
from poexy_core.manifest.manifest import PackageInfoManifest
from poexy_core.packages.format import PackageFormat
from poexy_core.pyproject.tables.poexy import Poexy

logger = logging.getLogger(__name__)


class SdistMetadata:
    def __init__(self, path: Path, name: str, version: str):
        self.path: Path = path
        self.name: str = distribution_name(NormalizedName(name))
        self.version: str = version
        self.archive_path: Path = self.__archive_path()
        self.root_folder: Path = self.__root_folder()

    def __archive_path(self) -> Path:
        return self.path / Path(f"{self.name}-{self.version}.tar.gz")

    def __root_folder(self) -> Path:
        return self.path / Path(f"{self.name}-{self.version}")


class SdistBuilder(Builder):
    def __init__(
        self,
        poetry: Poetry,
        poexy: Poexy,
        _format: PackageFormat,
        sdist_directory: Path,
        config_settings: dict[str, Any] | None = None,
    ):
        if _format != PackageFormat.Source:
            raise ValueError(f"Invalid format: {_format}")
        super().__init__(poetry, poexy, _format, sdist_directory, config_settings)
        self.__metadata = SdistMetadata(
            self.temp_destination_directory,
            self.poetry.package.name,
            self.poetry.package.version.to_string(),
        )
        self.__manifest = PackageInfoManifest(self.__metadata.root_folder)
        self.__archive: Optional[tarfile.TarFile] = None

        self._hooks.append(
            ReadmeHookBuilder(self.poetry, self.poexy, self.__metadata.root_folder)
        )
        self._hooks.append(
            PackageFilesHookBuilder(
                self.poexy, self.format, lambda path: self.__metadata.root_folder
            )
        )
        self._hooks.append(
            IncludeFilesHookBuilder(
                self.poexy, self.format, self.__metadata.root_folder
            )
        )
        self._hooks.append(LicenseHookBuilder(self.poexy, self.__metadata.root_folder))

    def __add_files(self):
        for hook in self._hooks:
            hook.add_files(self.__add_files_to_archive)

    @contextmanager
    def __create_archive(self) -> Generator[None, None, None]:
        logger.info(
            "Creating archive: "
            f"{self.__metadata.archive_path.relative_to(self.__metadata.path)}"
        )
        self.__archive = tarfile.open(
            self.__metadata.archive_path, "w:gz", format=tarfile.GNU_FORMAT
        )
        yield
        self.__archive.close()
        self.__archive = None

    def __add_files_to_archive(self, source: Path, destination: Path):
        if self.__archive is None:
            raise ValueError("Archive not created")
        relative_path = destination.relative_to(self.__metadata.root_folder.parent)
        logger.info(f"Adding: {relative_path}")
        self.__archive.add(source, arcname=relative_path)

    def __add_pkginfo_and_pyproject_to_archive(self):
        if self.__archive is None:
            raise ValueError("Archive not created")
        pyproject_path = self.poetry.pyproject.path
        self.__add_files_to_archive(
            pyproject_path,
            self.__metadata.root_folder / pyproject_path.name,
        )
        self.__add_files_to_archive(
            self.__manifest.path,
            self.__metadata.root_folder / self.__manifest.path.name,
        )

    @override
    @contextmanager
    def build(self) -> Generator[Path, None, None]:
        logger.info("Building sdist...")

        for hook in self._hooks:
            hook.build()

        with self.__create_archive():
            logger.info("Adding metadata...")
            self._add_metadata(self.__manifest)
            logger.info("Writing manifest...")
            self.__manifest.write()
            logger.info("Adding files...")
            self.__add_files()
            logger.info("Adding pkginfo and pyproject files...")
            self.__add_pkginfo_and_pyproject_to_archive()

        logger.info("Built successfully.")

        shutil.copy(self.__metadata.archive_path, self.destination_directory)

        yield self.__metadata.archive_path



import logging
import shutil
import sys
from pathlib import Path
from typing import Any, List, Optional

from poetry.core.masonry.metadata import Metadata
from poetry.core.poetry import Poetry

from poexy_core.builders.hooks.hook import HookBuilder
from poexy_core.manifest.manifest import Manifest
from poexy_core.metadata.builder import MetadataManifestBuilder
from poexy_core.metadata.fields import MetadataVersions
from poexy_core.packages.format import PackageFormat
from poexy_core.pyproject.tables.poexy import Poexy
from poexy_core.utils import python_impl

logger = logging.getLogger(__name__)


class PythonTag:
    def __init__(
        self,
        impl: Optional[str] = None,
        major: Optional[int] = None,
        minor: Optional[int] = None,
        abi: Optional[str] = None,
        platform: Optional[str] = None,
    ):
        self.impl: Optional[str] = impl
        self.major: Optional[int] = major
        self.minor: Optional[int] = minor
        self.platform: Optional[str] = platform
        self.abi: Optional[str] = abi

    @staticmethod
    def from_current_environment() -> "PythonTag":
        return PythonTag(
            impl=python_impl.get_python_implementation_tag(),
            major=sys.version_info.major,
            minor=sys.version_info.minor,
            abi=python_impl.get_abi_tag(),
            platform=python_impl.get_platform_tag(),
        )

    def __eq__(self, other: "PythonTag") -> bool:
        return str(self) == str(other)

    def __ne__(self, other: "PythonTag") -> bool:
        return not self.__eq__(other)

    def __hash__(self) -> int:
        return hash(str(self))

    def __str__(self) -> str:
        if self.impl is None:
            impl = "py"
        else:
            impl = self.impl

        if self.major is None:
            major = sys.version_info.major
        else:
            major = self.major

        if self.minor is None:
            minor = sys.version_info.minor
        else:
            minor = self.minor

        if self.minor is None:
            python_tag = f"{impl}{major}"
        else:
            python_tag = f"{impl}{major}{minor}"

        if self.abi is None:
            abi_tag = "none"
        else:
            abi_tag = self.abi

        if self.platform is None:
            platform_tag = "any"
        else:
            platform_tag = self.platform.replace("-", "_").replace(".", "_")

        tag = f"{python_tag}-{abi_tag}-{platform_tag}"

        return tag


class Builder:
    def __init__(
        self,
        poetry: Poetry,
        poexy: Poexy,
        _format: PackageFormat,
        destination_directory: Path,
        config_settings: dict[str, Any] | None = None,
    ):
        self.poetry = poetry
        self.poexy = poexy
        self.format = _format
        self.destination_directory = destination_directory
        self.temp_destination_directory = destination_directory / "temp"
        self.config_settings = config_settings
        self.package_name = self.poetry.package.name
        self.package_version = self.poetry.package.version

        shutil.rmtree(self.temp_destination_directory, ignore_errors=True)

        if not self.destination_directory.exists():
            self.destination_directory.mkdir(parents=True, exist_ok=True)

        if not self.temp_destination_directory.exists():
            self.temp_destination_directory.mkdir(parents=True, exist_ok=True)

        self._hooks: List[HookBuilder] = []

    def add_hook(self, hook: HookBuilder):
        self._hooks.append(hook)

    def _add_metadata(self, manifest: Manifest) -> MetadataManifestBuilder:
        if self.poexy.readme is not None:
            self.poexy.readme.transform_poetry_package(self.poetry.package)
        metadata = Metadata.from_package(self.poetry.package)
        for hook in self._hooks:
            hook.add_metadata(metadata, self.format)
        builder = MetadataManifestBuilder(manifest, MetadataVersions.V2_4)
        MetadataManifestBuilder.from_poetry(builder, metadata)
        return builder

    def build(self) -> Path:
        raise NotImplementedError("Subclass must implement build method")



from enum import Enum


class PackageFormat(str, Enum):
    Source = "source"
    Wheel = "wheel"
    Binary = "binary"


DEFAULT_FORMATS = [PackageFormat.Source, PackageFormat.Wheel, PackageFormat.Binary]


class WheelFormat(str, Enum):
    Source = "source"
    Binary = "binary"


DEFAULT_WHEEL_FORMATS = [WheelFormat.Source]



import sys
from pathlib import Path
from typing import Callable

import pytest
from assertpy import assert_that
from poetry.core.masonry.utils.helpers import distribution_name

from poexy_core.builders.builder import PythonTag
from poexy_core.builders.sdist import SdistMetadata
from poexy_core.builders.wheel import WheelMetadata

# pylint: disable=redefined-outer-name


@pytest.fixture()
def default_python_tag():
    return PythonTag(impl="py", major=sys.version_info.major)


@pytest.fixture()
def current_python_tag():
    return PythonTag.from_current_environment()


@pytest.fixture()
def package_name(pyproject) -> Callable[[], str]:
    def _package_name():
        return pyproject().poetry.package.name

    return _package_name


@pytest.fixture()
def dist_package_name(pyproject) -> Callable[[], str]:
    def _dist_package_name():
        return distribution_name(pyproject().poetry.package.name)

    return _dist_package_name


@pytest.fixture()
def package_version(pyproject) -> Callable[[], str]:
    def _package_version():
        return pyproject().poetry.package.version.to_string()

    return _package_version


@pytest.fixture()
def wheel_metadata(
    dist_temp_path, dist_package_name, package_version
) -> Callable[[PythonTag], WheelMetadata]:
    def _wheel_metadata(python_tag: PythonTag) -> WheelMetadata:
        metadata = WheelMetadata(
            dist_temp_path, dist_package_name(), package_version(), python_tag
        )
        assert_that(str(metadata.archive_path)).exists()
        return metadata

    return _wheel_metadata


@pytest.fixture()
def wheel_dist_info_folder(wheel_metadata) -> Callable[[PythonTag], Path]:
    def _wheel_dist_info_folder(python_tag: PythonTag):
        metadata = wheel_metadata(python_tag)
        dist_info_folder = metadata.dist_info_folder
        dist_info_folder = dist_info_folder.relative_to(metadata.root_folder)
        return dist_info_folder

    return _wheel_dist_info_folder


@pytest.fixture()
def wheel_data_purelib_folder(wheel_metadata) -> Callable[[PythonTag], Path]:
    def _wheel_data_purelib_folder(python_tag: PythonTag):
        metadata = wheel_metadata(python_tag)
        data_purelib_folder = metadata.data_purelib_folder
        data_purelib_folder = data_purelib_folder.relative_to(metadata.root_folder)
        return data_purelib_folder

    return _wheel_data_purelib_folder


@pytest.fixture()
def wheel_data_scripts_folder(wheel_metadata) -> Callable[[PythonTag], Path]:
    def _wheel_data_scripts_folder(python_tag: PythonTag):
        metadata = wheel_metadata(python_tag)
        data_scripts_folder = metadata.data_scripts_folder
        data_scripts_folder = data_scripts_folder.relative_to(metadata.root_folder)
        return data_scripts_folder

    return _wheel_data_scripts_folder


@pytest.fixture()
def sdist_metadata(
    dist_temp_path, dist_package_name, package_version
) -> Callable[[], SdistMetadata]:
    def _sdist_metadata():
        metadata = SdistMetadata(dist_temp_path, dist_package_name(), package_version())
        assert_that(str(metadata.archive_path)).exists()
        return metadata

    return _sdist_metadata



from typing import override

from poetry.core.masonry.metadata import Metadata

from poexy_core.builders.binary import BinaryBuilder
from poexy_core.builders.hooks.hook import HookBuilder
from poexy_core.builders.types import FilePathCallback
from poexy_core.packages.format import PackageFormat


class BinaryHookBuilder(HookBuilder):
    def __init__(self, builder: BinaryBuilder):
        super().__init__("binary")
        self.__builder = builder
        self.__destination = builder._metadata.data_scripts_folder
        self.__files = []

    @override
    def build(self):
        with self._hook_build():
            source, destination = self.__builder.build_executable()
            self.__files.append((source, destination))

    @override
    def add_files(self, callback: FilePathCallback):
        with self._hook_add_files():
            for source, destination in self.__files:
                callback(source, self.__destination / destination)

    @override
    def add_metadata(self, metadata: Metadata, _format: PackageFormat):
        pass



import pytest

# pylint: disable=redefined-outer-name


@pytest.fixture()
def project_path(sample_project):
    return sample_project("readme-undetermined-suffix")


def test_wheel(project, project_path, assert_wheel_build):
    with project(project_path):
        with pytest.raises(
            ValueError,
            match="Cannot determine content type from file extension: .txt",
        ):
            assert_wheel_build(project_path)


def test_sdist(project, project_path, assert_sdist_build):
    with project(project_path):
        with pytest.raises(
            ValueError,
            match="Cannot determine content type from file extension: .txt",
        ):
            assert_sdist_build(project_path)
